# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pickle

import cv2
import numpy as np

from agml.framework import AgMLSerializable
from agml.utils.logging import log
from agml.utils.image import imread_context
from agml.utils.general import resolve_tuple
from agml.utils.io import recursive_dirname


class ImageResizeManager(AgMLSerializable):
    """Smartly resizes input image and annotation data.

    This class wraps the resizing of image and annotation data and
    dynamically resizes images and annotations based on the user
    preference. Resizing is applies to input images and annotations
    where required, e.g. masks for semantic segmentation, or COCO
    JSON bounding boxes for object detection.

    Image resizing contains a few modes:

    1. `default`: No resizing, leaves the images in their default size.
       This is the default parameter if nothing is passed.
    2. `train`: This will set a default training size of (512, 512).
    3. `imagenet`: This will set a default size of (224, 224).
    4. custom size: Resizes the images to the provided size.
    5. `auto`: Dynamically selects an image size based on a few factors.
       For example, if there are certain outliers in a dataset which are
       of a different size while the majority remain the same, then, the
       behavior of this method is chosen by a majority threshold of the
       sizes of all the images in the dataset. If no shape can be inferenced,
       it returns a default size of (512, 512). The logic for the `auto`
       case is explained in more detail below.

    The actual resizing caveats are as follows:

    - For object detection, the bounding box coordinates will
      be resized and the area of the box will in turn be recomputed.
    - For semantic segmentation, the annotation mask will be resized,
      using a nearest-neighbor interpolation to keep it as similar
      as possible to the original mask (preventing data loss).

    Note: Images will be resized *before* being passed into
    a transformation pipeline. This is to prevent data loss.
    """
    serializable = frozenset(
        ('task', 'dataset_name', 'dataset_root', 'auto_enabled',
         'resize_type', 'image_size', 'interpolation'))

    # Stores the path to the local file which contains the
    # information on the image shapes in all of the datasets.
    _shape_info_file = os.path.join(
        recursive_dirname(__file__, 3),
        '_assets', 'shape_info.pickle')

    # Stores the default image size if it is not inferenced.
    _default_size = (512, 512)

    # Stores the default imagenet model input shape.
    _imagenet_size = (224, 224)

    def __init__(self, task, dataset, root):
        self._task = task
        self._dataset_name = dataset
        self._dataset_root = root

        self._resize_type = 'default'
        self._image_size = None
        self._interpolation = cv2.INTER_LINEAR

        self._auto_enabled = True

    def disable_auto(self):
        # For multi-dataset loaders, the `auto` option is disabled.
        self._auto_enabled = False

    @property
    def state(self):
        return self._resize_type

    @property
    def size(self):
        return self._image_size

    @staticmethod
    def _tuple_euclidean(t1, t2):
        (x1, y1), (x2, y2) = t1, t2
        return np.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)

    def _method_resize(self, image, size):
        return cv2.resize(image, size, interpolation = self._interpolation)

    def assign(self, kind, method = None):
        """Assigns the resize parameter (and does necessary calculations)."""
        if kind == 'default':
            self._resize_type = 'default'
            self._image_size = None
        elif kind == 'train':
            self._resize_type = 'train'
            self._image_size = self._default_size
        elif kind == 'imagenet':
            self._resize_type = 'imagenet'
            self._image_size = self._imagenet_size
        elif isinstance(kind, (list, tuple, np.ndarray, set)):
            if not len(kind) == 2:
                raise ValueError(
                    f"Got a sequence {kind}, expected two values for "
                    f"the height and width but got {len(kind)} values.")
            self._resize_type = 'custom_size'
            self._image_size = tuple([i for i in kind])
        elif 'auto' in kind:
            # This means that the dataloader has been exported in some format.
            # First, we check whether a different size has automatically been
            # set, since if it has, we don't want to override that size.
            if kind == 'train-auto':
                if not self._auto_enabled:
                    self._resize_type = 'train'
                    self._image_size = self._default_size
                elif self._resize_type == 'default':
                    info = self._maybe_load_shape_info()
                    if info is None:
                        shape = self._random_inference_shape()
                    else:
                        shape = self._inference_shape(info)
                    self._resize_type = 'auto'
                    self._image_size = resolve_tuple(shape)

        # Check interpolation method independently.
        if method is not None:
            method = method.lower()
            if method not in ['bilinear', 'area', 'nearest', 'cubic']:
                raise ValueError(f"Invalid interpolation method: {method}.")
            self._interpolation = {
                'bilinear': cv2.INTER_LINEAR,
                'area': cv2.INTER_AREA,
                'nearest': cv2.INTER_NEAREST,
                'cubic': cv2.INTER_CUBIC
            }[method]

    def apply(self, contents):
        """Applies the resizing operation to the input data."""
        if self._task in ['image_classification', 'image_regression']:
            return self._resize_image_input(
                contents, self._image_size)
        elif self._task == 'object_detection':
            return self._resize_image_and_coco(
                contents, self._image_size)
        elif self._task == 'semantic_segmentation':
            return self._resize_image_and_mask(
                contents, self._image_size)

    def _inference_shape(self, info):
        """Attempts to inference a shape for the `auto` method.

        The logic for this method goes in the following order:

        1. If there is only one unique shape in the dataset, then that
           is the shape that is used.
        2. If one shape makes up more than 70% of the shapes in the
           dataset, (7/10 of the dataset), then that shape is used.
        3. If there are less than or equal to 4 shapes which make up
           at least 10% of the dataset each (e.g., there may be 40%
           one shape, 40% another shape, and then 20% a third and final
           shape), then the shape closest to (512, 512) will be used. If
           there is one that makes up a majority, however, so greater
           than 50% of the dataset, and the next-largest is less than 30%,
           then that majority size will be returned.

        4. Otherwise, no specific shape can be inferenced, so the default
           shape of (512, 512) is used instead.

        Note that the `auto` method is very experimental, and is more useful
        just for instantly determining when a dataset has only one type
        of shape rather than having to manually figure sizes out. If you
        encounter issues in training when there are multiple pluralities
        of shapes and no explicitly majority, then do not use this method.
        """
        shapes, counts = info
        shapes = shapes[:, :-1]
        if len(shapes) == 1:
            return shapes[0].tolist()

        total_count = np.sum(counts)
        if np.max(counts) > int(total_count * 0.7):
            return shapes[np.where(counts == np.max(counts))[0]][0].tolist()

        if len(shapes) <= 4:
            count_proportions = counts.copy() / total_count
            maybe_max = np.where(count_proportions >= 0.5)[0]
            if maybe_max:
                if maybe_max.copy().sort()[-2] <= 0.3:
                    return shapes[maybe_max].tolist()

            two_pluralities = shapes[np.where(count_proportions >= 0.35)[0]]
            if len(two_pluralities) != 2:
                return self._get_log_default_shape()
            outs = (self._tuple_euclidean(two_pluralities[0], self._default_size),
                    self._tuple_euclidean(two_pluralities[1], self._default_size))
            return two_pluralities[outs.index(max(outs))][0].tolist()

        return self._get_log_default_shape()

    def _random_inference_shape(self):
        """Randomly attempts to inference a shape for the dataset.

        In the case that there is no shape information saved in the
        local `.shape.info.pickle` file, this method randomly attempts
        to inference a dataset shape by randomly selecting images from
        the dataset and performing calculations (similarly to what it
        would do in the regular shape inference method, only with a
        random sample instead of the entire dataset for reference.
        """
        log(f"Could not find shape information for {self._dataset_name}. "
            f"Attempting to randomly inference the dataset shape.")

        image_path = os.path.join(self._dataset_root, 'images')
        images = np.random.choice(os.listdir(image_path), size = 25)

        # Get all of the shapes from the random sample of images.
        shapes = []
        for path in images:
            path = os.path.join(image_path, path)
            with imread_context(path) as image:
                shapes.append(image)

        # Inference a valid shape from the shapes. We dispatch to the
        # regular inferencing method once we have the shapes and counts.
        unique_shapes, counts = np.unique(shapes, return_counts = True, axis = 0)
        return self._inference_shape((unique_shapes, counts))

    def _maybe_load_shape_info(self):
        """Loads the contents of the shape information file."""
        try:
            with open(self._shape_info_file, 'rb') as f:
                contents = pickle.load(f)
                return contents.get(self._dataset_name, None)
        except OSError:
            raise EnvironmentError(
                f"Could not find the local file {self._shape_info_file} "
                f"containing shape information. There was likely a problem "
                f"when building AgML. Please re-install it.")

    def _get_log_default_shape(self):
        """Returns the default shape and warns that no shape could be inferenced."""
        log(f"No specific shape could be inferenced for the dataset "
            f"{self._dataset_name}. Using the default shape of (512, "
            f"512). Please use this instead of 'auto'. You can set this "
            f"shape by default by using the parameter 'train'.")
        return self._default_size

    # The following methods conduct the actual resizing of the images
    # and potentially their annotations for the different tasks.

    def _resize_image_input(self, contents, image_size):
        # If there is only one image input, just return that.
        image, label = contents
        if isinstance(image, np.ndarray):
            return self._resize_single_image(contents, image_size)
        if image_size is not None:
            return {
                k: self._method_resize(
                    i.astype(np.uint16), image_size).astype(np.int32)
                for k, i in image.items()}, label
        return image, label

    def _resize_single_image(self, contents, image_size):
        image, label = contents
        if image_size is not None:
            # No processing done on the annotation.
            image = self._method_resize(image, image_size)
        return image, label

    def _resize_image_and_coco(self, contents, image_size):
        image, coco = contents
        if image_size is not None:
            # Extract the original size of the bounding boxes.
            y_scale, x_scale = image.shape[0:2]
            bboxes = coco['bbox']
            processed_bboxes = []
            for bbox in bboxes:
                x1, y1, width, height = bbox
                processed_bboxes.append([
                    x1 / x_scale, y1 / y_scale,
                    width / x_scale, height / y_scale])

            # Clip the bounding boxes (There might be a case where
            # the bounding box is on the edge, but goes to just over 1.0
            # which in turn causes many bugs. This prevents that).
            processed_bboxes = np.array(processed_bboxes)
            np.clip(processed_bboxes, 0, 1, processed_bboxes)

            # Resize the image and calculate its new size ratio.
            image = self._method_resize(image, image_size)
            y_new, x_new = image.shape[0:2]

            # Update the bounding boxes and areas with the new ratio.
            final_processed_bboxes, areas = [], []
            for bbox in processed_bboxes:
                x1, y1, width, height = bbox
                new_bbox = [
                    int(x1 * x_new), int(y1 * y_new),
                    int(width * x_new), int(height * y_new)]
                final_processed_bboxes.append(new_bbox)
                areas.append(new_bbox[2] * new_bbox[3])
            final_processed_bboxes = \
                np.array(final_processed_bboxes).astype(np.int32)
            coco['bbox'] = final_processed_bboxes
            areas = np.array(areas).astype(np.int32)
            coco['area'] = areas
        return image, coco

    def _resize_image_and_mask(self, contents, image_size):
        image, mask = contents
        if image_size is not None:
            # Resize the image and the mask together if requested.
            image = self._method_resize(image, image_size)
            mask = cv2.resize(mask, image_size, cv2.INTER_NEAREST)
        return image, mask









# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from tqdm import tqdm

import torch
import torch.nn as nn
from torchvision.models import efficientnet_b4

from agml.models.base import AgMLModelBase
from agml.models.tools import auto_move_data, imagenet_style_process
from agml.models.metrics.accuracy import Accuracy
from agml.data.public import source


class EfficientNetB4Transfer(nn.Module):
    """Wraps an EfficientDetB4 model with a classification head."""
    def __init__(self, num_classes):
        super(EfficientNetB4Transfer, self).__init__()
        self.base = efficientnet_b4(pretrained = False)
        self.l1 = nn.Linear(1000, 256)
        self.dropout = nn.Dropout(0.1)
        self.relu = nn.ReLU()
        self.l2 = nn.Linear(256, num_classes)

    def forward(self, x, **kwargs): # noqa
        x = self.base(x)
        x = x.view(x.size(0), -1)
        x = self.dropout(self.relu(self.l1(x)))
        return self.l2(x)


class ClassificationModel(AgMLModelBase):
    """Wraps an `EfficientNetB4` model for agricultural image classification.

    When using the model for inference, you should use the `predict()` method
    on any set of input images. This method wraps the `forward()` call with
    additional steps which perform the necessary preprocessing on the inputs,
    including resizing, normalization, and batching, as well as additional
    post-processing on the outputs (such as converting one-hot labels to
    integer labels), which will allow for a streamlined inference pipeline.

    If you want to use the `forward()` method directly, so that you can just
    call `model(inputs)`, then make sure your inputs are properly processed.
    This can be done by calling `model.preprocess_input(inputs)` on the
    input list of images/single image and then passing that result to the model.
    This will also return a one-hot label feature vector instead of integer
    labels, in the case that you want further customization of the outputs.

    This model can be subclassed in order to run a full training job; the
    actual transfer `EfficientNetB4` model can be accessed through the
    parameter `net`, and you'll need to implement methods like `training_step`,
    `configure_optimizers`, etc. See PyTorch Lightning for more information.
    """
    serializable = frozenset(("model", ))
    state_override = serializable

    def __init__(self, dataset):
        # Construct the network and load in pretrained weights.
        super(ClassificationModel, self).__init__()
        self.net = self._construct_sub_net(dataset)

    @auto_move_data
    def forward(self, batch):
        return self.net(batch)

    @staticmethod
    def _construct_sub_net(dataset):
        return EfficientNetB4Transfer(source(dataset).num_classes)

    @staticmethod
    def _preprocess_image(image):
        """Preprocesses a single input image to EfficientNet standards.

        The preprocessing steps are applied logically; if the images
        are passed with preprocessing already having been applied, for
        instance, the images are already resized or they are already been
        normalized, the operation is not applied again, for efficiency.

        Preprocessing includes the following steps:

        1. Resizing the image to size (224, 224).
        2. Performing normalization with ImageNet parameters.
        3. Converting the image into a PyTorch tensor format.

        as well as other intermediate steps such as adding a channel
        dimension for two-channel inputs, for example.
        """
        return imagenet_style_process(image)

    @staticmethod
    def preprocess_input(images = None) -> "torch.Tensor":
        """Preprocesses the input image to the specification of the model.

        This method takes in a set of inputs and preprocesses them into the
        expected format for the `EfficientNetB4` image classification model.
        There are a variety of inputs which are accepted, including images,
        image paths, as well as fully-processed image tensors. The inputs
        are expanded and standardized, then run through a preprocessing
        pipeline which formats them into a single tensor ready for the model.

        Preprocessing steps include normalization, resizing, and converting
        to the channels-first format used by PyTorch models. The output
        of this method will be a single `torch.Tensor`, which has shape
        [N, C, H, W], where `N` is the batch dimension. If only a single
        image is passed, this will have a value of 1.

        This method is largely beneficial when you just want to preprocess
        images into the specification of the model, without getting the output.
        Namely, `predict()` is essentially just a wrapper around this method
        and `forward()`, so you can run this externally and then run `forward()`
        to get the original model outputs, without any extra post-processing.

        Parameters
        ----------
        images : Any
            One of the following formats (and types):
                1. A single image path (str)
                2. A list of image paths (List[str])
                3. A single image (np.ndarray, torch.Tensor)
                4. A list of images (List[np.ndarray, torch.Tensor])
                5. A batched tensor of images (np.ndarray, torch.Tensor)

        Returns
        -------
        A 4-dimensional, preprocessed `torch.Tensor`.
        """
        images = ClassificationModel._expand_input_images(images)
        return torch.stack(
            [ClassificationModel._preprocess_image(
                image) for image in images], dim = 0)

    @torch.no_grad()
    def predict(self, images):
        """Runs `EfficientNetB4` inference on the input image(s).

        This method is the primary inference method for the model; it
        accepts a set of input images (see `preprocess_input()` for a
        detailed specification on the allowed input parameters), then
        preprocesses them to the model specifications, forward passes
        them through the model, and finally returns the predictions.

        In essence, this method is a wrapper for `forward()` that allows
        for passing a variety of inputs. If, on the other hand, you
        have pre-processed inputs and simply want to forward pass through
        the model without having to spend computational time on what
        is now unnecessary preprocessing, you can simply call `forward()`
        and then run `torch.argmax()` on the outputs to get predictions.

        Parameters
        ----------
        images : Any
            See `preprocess_input()` for the allowed input images.

        Returns
        -------
        A `np.ndarray` with integer labels for each image.
        """
        images = self.preprocess_input(images)
        return self._to_out(torch.squeeze(torch.argmax(self.forward(images), 1)))

    def evaluate(self, loader, **kwargs):
        """Runs an accuracy evaluation on the given loader.

        This method will loop over the `AgMLDataLoader` and compute accuracy.

        Parameters
        ----------
        loader : AgMLDataLoader
            A semantic segmentation loader with the dataset you want to evaluate.

        Returns
        -------
        The final calculated accuracy.
        """
        # Construct the metric and run the calculations.
        acc = Accuracy()
        bar = tqdm(loader, desc = "Calculating Accuracy")
        for sample in bar:
            image, truth = sample
            pred_label = self.predict(image)
            acc.update([pred_label], [truth])
            bar.set_postfix({'accuracy': acc.compute().numpy().item()})

        # Compute the final accuracy.
        return acc.compute().numpy().item()







# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable
from functools import wraps

import torch
import numpy as np
import albumentations as A

from agml.backend.tftorch import is_array_like


def imagenet_style_process(image, size = None):
    """Preprocesses a single input image to ImageNet standards.

    The preprocessing steps are applied logically; if the images
    are passed with preprocessing already having been applied, for
    instance, the images are already resized or they are already been
    normalized, the operation is not applied again, for efficiency.

    Preprocessing includes the following steps:

    1. Resizing the image to size (224, 224).
    2. Performing normalization with ImageNet parameters.
    3. Converting the image into a PyTorch tensor format.

    as well as other intermediate steps such as adding a channel
    dimension for two-channel inputs, for example.
    """
    # Convert the image to a NumPy array.
    if is_array_like(image) and hasattr(image, 'numpy'):
        image = image.numpy()

    # Add a channel dimension for grayscale imagery.
    if image.ndim == 2:
        image = np.expand_dims(image, axis = -1)

    # If the image is already in channels-first format, convert
    # it back temporarily until preprocessing has concluded.
    if image.shape[0] <= 3:
        image = np.transpose(image, (1, 2, 0))

    # Resize the image to ImageNet standards.
    h = w = 224
    if size is not None:
        h, w = size
    rz = A.Resize(height = h, width = w)
    if image.shape[0] != h or image.shape[1] != w:
        image = rz(image = image)['image']

    # Normalize the image to ImageNet standards.
    if 1 <= image.max() <= 255:
        mean = [0.485, 0.456, 0.406]
        std = [0.229, 0.224, 0.225]
        image = image.astype(np.float32) / 255.
        mean = np.array(mean, dtype = np.float32)
        std = np.array(std, dtype = np.float32)
        denominator = np.reciprocal(std, dtype = np.float32)
        image = (image - mean) * denominator

    # Convert the image into a PyTorch tensor.
    image = torch.from_numpy(image).permute(2, 0, 1)

    # Return the processed image.
    return image


# Ported from PyTorch Lightning v1.3.0.
def auto_move_data(fn: Callable) -> Callable:
    """
    Decorator for :class:`~pytorch_lightning.core.lightning.LightningModule` methods for which
    input arguments should be moved automatically to the correct device.
    It has no effect if applied to a method of an object that is not an instance of
    :class:`~pytorch_lightning.core.lightning.LightningModule` and is typically applied to ``__call__``
    or ``forward``.

    Args:
        fn: A LightningModule method for which the arguments should be moved to the device
            the parameters are on.

    Example::

        # directly in the source code
        class LitModel(LightningModule):

            @auto_move_data
            def forward(self, x):
                return x

        # or outside
        LitModel.forward = auto_move_data(LitModel.forward)

        model = LitModel()
        model = model.to('cuda')
        model(torch.zeros(1, 3))

        # input gets moved to device
        # tensor([[0., 0., 0.]], device='cuda:0')

    """

    @wraps(fn)
    def auto_transfer_args(self, *args, **kwargs):
        from pytorch_lightning import LightningModule

        if not isinstance(self, LightningModule):
            return fn(self, *args, **kwargs)

        args, kwargs = self.transfer_batch_to_device(
            (args, kwargs), device = self.device, dataloader_idx = None) # noqa
        return fn(self, *args, **kwargs)

    return auto_transfer_args





# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

import torch
import numpy as np
import albumentations as A

from tqdm import tqdm
from ensemble_boxes.ensemble_boxes_wbf import weighted_boxes_fusion

from effdet import (
    create_model_from_config,
    get_efficientdet_config,
    DetBenchPredict,
    DetBenchTrain
)

from agml.models.base import AgMLModelBase
from agml.models.benchmarks import BenchmarkMetadata
from agml.models.tools import auto_move_data
from agml.models.metrics.map import MeanAveragePrecision
from agml.data.public import source
from agml.backend.tftorch import is_array_like
from agml.utils.image import resolve_image_size
from agml.utils.logging import log
from agml.viz.boxes import visualize_image_and_boxes


class DetectionModel(AgMLModelBase):
    """Wraps an `EfficientDetD4` model for agricultural object detection.

    When using the model for inference, you should use the `predict()` method
    on any set of input images. This method wraps the `forward()` call with
    additional steps which perform the necessary preprocessing on the inputs,
    including resizing, normalization, and batching, as well as additional
    post-processing on the outputs (such as converting one-hot labels to
    integer labels), which will allow for a streamlined inference pipeline.

    If you want to use the `forward()` method directly, so that you can just
    call `model(inputs)`, then make sure your inputs are properly processed.
    This can be done by calling `model.preprocess_input(inputs)` on the
    input list of images/single image and then passing that result to the model.
    This will also return a one-hot label feature vector instead of integer
    labels, in the case that you want further customization of the outputs.

    By default, when instantiating a `DetectionModel`, it is prepared in
    inference mode. In order to use this model for training, you need to convert
    it to training mode by using `DetectionModel.switch_train()`. To convert
    back to inference mode, use `DetectionModel.switch_predict()`.

    If you want to use your own custom model and/or training pipeline, without the
    existing input restrictions, then you can subclass this model. In the `super`
    call in the `__init__` method, pass the parameter `model_initialized = True`,
    which will enable you to initialize the model in your own format.

    Parameters
    ----------
    num_classes : int
        The number of classes for the `EfficientDet` model.
    image_size : int, tuple
        The shape of image inputs to the model.
    conf_threshold : float
        Filters bounding boxes by their level of confidence based on this threshold.
    """
    serializable = frozenset((
        "model", "num_classes", "conf_thresh", "image_size"))
    state_override = frozenset(("model",))

    def __init__(self,
                 num_classes = 1,
                 image_size = 512,
                 conf_threshold = 0.3,
                 **kwargs):
        # Initialize the base modules.
        super(DetectionModel, self).__init__()

        # If being initialized by a subclass, then don't do any of
        # model construction logic (since that's already been done).
        if not kwargs.get('model_initialized', False):
            # Construct the network and load in pretrained weights.
            self._image_size = resolve_image_size(image_size)
            self._confidence_threshold = conf_threshold
            self._num_classes = num_classes
            self.model = self._construct_sub_net(
                self._num_classes, self._image_size)

        # Filter out unnecessary warnings.
        warnings.filterwarnings(
            'ignore', category = UserWarning, module = 'ensemble_boxes')
        warnings.filterwarnings(
            'ignore', category = UserWarning, module = 'effdet.bench')  # noqa

    @auto_move_data
    def forward(self, batch):
        """Ensures that the input is valid for the model."""
        return self.model(batch)

    @staticmethod
    def _construct_sub_net(num_classes, image_size):
        cfg = get_efficientdet_config('tf_efficientdet_d4')
        cfg.update({"image_size": image_size})
        model = create_model_from_config(
            cfg, pretrained = False,
            num_classes = num_classes)
        return DetBenchPredict(model)

    def switch_predict(self):
        """Prepares the model for evaluation mode."""
        state = self.model.state_dict()
        if not isinstance(self.model, DetBenchPredict):
            self.model = DetBenchPredict(self.model.model)
        self.model.load_state_dict(state)

    def switch_train(self):
        """Prepares the model for training mode."""
        state = self.model.state_dict()
        if not isinstance(self.model, DetBenchTrain):
            self.model = DetBenchTrain(self.model.model)
        self.model.load_state_dict(state)

    @property
    def original(self):  # override for detection models.
        return self.model.model

    @torch.jit.ignore()
    def reset_class_net(self, num_classes = 1):
        """Reconfigures the output class net for a new number of classes.

        Parameters
        ----------
        num_classes : int
            The number of classes to reconfigure the output net to use.
        """
        if num_classes != self._num_classes:
            self.model.model.reset_head(num_classes = num_classes)

    @staticmethod
    def _preprocess_image(image, image_size):
        """Preprocesses a single input image to EfficientNet standards.

        The preprocessing steps are applied logically; if the images
        are passed with preprocessing already having been applied, for
        instance, the images are already resized or they are already been
        normalized, the operation is not applied again, for efficiency.

        Preprocessing includes the following steps:

        1. Resizing the image to size (224, 224).
        2. Performing normalization with ImageNet parameters.
        3. Converting the image into a PyTorch tensor format.

        as well as other intermediate steps such as adding a channel
        dimension for two-channel inputs, for example.
        """
        # Convert the image to a NumPy array.
        if is_array_like(image) and hasattr(image, 'numpy'):
            image = image.numpy()

        # Add a channel dimension for grayscale imagery.
        if image.ndim == 2:
            image = np.expand_dims(image, axis = -1)

        # If the image is already in channels-first format, convert
        # it back temporarily until preprocessing has concluded.
        if image.shape[0] <= 3:
            image = np.transpose(image, (1, 2, 0))

        # Resize the image to ImageNet standards.
        (w, h) = image_size
        rz = A.Resize(height = h, width = w)
        if image.shape[0] != h or image.shape[1] != w:
            image = rz(image = image)['image']

        # Normalize the image to ImageNet standards.
        if 1 <= image.max() <= 255:
            image = image.astype(np.float32) / 255.

        # Convert the image into a PyTorch tensor.
        image = torch.from_numpy(image).permute(2, 0, 1)

        # Return the processed image.
        return image

    @torch.jit.ignore()
    def preprocess_input(self, images, return_shapes = False):
        """Preprocesses the input image to the specification of the model.

        This method takes in a set of inputs and preprocesses them into the
        expected format for the `EfficientDetD4` object detection model.
        There are a variety of inputs which are accepted, including images,
        image paths, as well as fully-processed image tensors. The inputs
        are expanded and standardized, then run through a preprocessing
        pipeline which formats them into a single tensor ready for the model.

        Preprocessing steps include normalization, resizing, and converting
        to the channels-first format used by PyTorch models. The output
        of this method will be a single `torch.Tensor`, which has shape
        [N, C, H, W], where `N` is the batch dimension. If only a single
        image is passed, this will have a value of 1.

        This method is largely beneficial when you just want to preprocess
        images into the specification of the model, without getting the output.
        Namely, `predict()` is essentially just a wrapper around this method
        and `forward()`, so you can run this externally and then run `forward()`
        to get the original model outputs, without any extra post-processing.

        Parameters
        ----------
        images : Any
            One of the following formats (and types):
                1. A single image path (str)
                2. A list of image paths (List[str])
                3. A single image (np.ndarray, torch.Tensor)
                4. A list of images (List[np.ndarray, torch.Tensor])
                5. A batched tensor of images (np.ndarray, torch.Tensor)
        return_shapes : bool
            Whether to return the original shapes of the input images.

        Returns
        -------
        A 4-dimensional, preprocessed `torch.Tensor`. If `return_shapes`
        is set to True, it also returns the original shapes of the images.
        """
        images = self._expand_input_images(images)
        shapes = self._get_shapes(images)
        images = torch.stack(
            [self._preprocess_image(
                image, self._image_size) for image in images], dim = 0)
        if return_shapes:
            return images, shapes
        return images

    def _process_detections(self, detections):
        """Post-processes the output detections (boxes, labels) from the model."""
        predictions = []

        # Convert all of the output detections into predictions. This involves
        # selecting the bounding boxes, confidence scores, and classes, and then
        # dropping any of them which do not have a confidence score about the
        # threshold as determined when the class is initialized.
        for d in detections:
            # Extract the bounding boxes, confidence scores,
            # and class labels from the output detections.
            boxes, scores, classes = d[:, :4], d[:, 4], d[:, 5]

            # Only return boxes which are above the confidence threshold.
            valid_indexes = np.where(scores > self._confidence_threshold)[0]
            boxes = boxes[valid_indexes]
            scores = scores[valid_indexes]
            classes = classes[valid_indexes]
            predictions.append({"boxes": boxes, "scores": scores, "classes": classes})

        # Run weighted boxes fusion. For an exact description of how this
        # works, see the paper: https://arxiv.org/pdf/1910.13302.pdf.
        (predicted_bboxes,
         predicted_class_confidences,
         predicted_class_labels) = self._wbf(predictions)
        return predicted_bboxes, predicted_class_confidences, predicted_class_labels

    @staticmethod
    def _rescale_bboxes(predicted_bboxes, image_sizes):
        """Re-scales output bounding boxes to the original image sizes."""
        scaled_boxes = []
        for bboxes, img_dims in zip(predicted_bboxes, image_sizes):
            h, w = img_dims
            if len(bboxes) > 0:
                # Re-scale the bounding box to the appropriate format.
                scale_ratio = [w / 512, h / 512, w / 512, h / 512]
                scaled = (np.array(bboxes) * scale_ratio).astype(np.int32)

                # Convert the Pascal-VOC (xyxy) format to COCO (xywh).
                x, y = scaled[:, 0], scaled[:, 1]
                w, h = scaled[:, 2] - x, scaled[:, 3] - y
                scaled_boxes.append(np.dstack((x, y, w, h)))
                continue

            # Otherwise, there is no prediction for this image.
            scaled_boxes.append(np.array([]))
        return scaled_boxes

    @staticmethod
    def _wbf(predictions):
        """Runs weighted boxes fusion on the output predictions."""
        bboxes, confidences, class_labels = [], [], []

        # Fuse the predictions for each in the batch.
        for prediction in predictions:
            boxes = [(prediction["boxes"] / 512).tolist()]
            scores = [prediction["scores"].tolist()]
            labels = [prediction["classes"].tolist()]

            # Run the actual fusion and update the containers.
            boxes, scores, labels = weighted_boxes_fusion(
                boxes, scores, labels,
                iou_thr = 0.44, skip_box_thr = 0.43)
            boxes = boxes * (512 - 1)
            bboxes.append(boxes)
            confidences.append(scores)
            class_labels.append(labels.astype(np.int32))
        return bboxes, confidences, class_labels

    @staticmethod
    def _remap_outputs(boxes, labels, confidences):
        """Remaps the outputs to the format described in `predict()`."""
        squeeze = lambda *args: tuple(list(
            np.squeeze(a).tolist() for a in args))
        return [squeeze(b, l, c)
                for b, l, c in zip(boxes, labels, confidences)]

    def _to_out(self, tensor: "torch.Tensor") -> "torch.Tensor":
        if isinstance(tensor, dict):
            tensor = tensor['detections']
        return super()._to_out(tensor)

    @torch.no_grad()
    def predict(self, images):
        """Runs `EfficientNetD4` inference on the input image(s).

        This method is the primary inference method for the model; it
        accepts a set of input images (see `preprocess_input()` for a
        detailed specification on the allowed input parameters), then
        preprocesses them to the model specifications, forward passes
        them through the model, and finally returns the predictions.

        In essence, this method is a wrapper for `forward()` that allows
        for passing a variety of inputs. If, on the other hand, you
        have pre-processed inputs and simply want to forward pass through
        the model without having to spend computational time on what
        is now unnecessary preprocessing, you can simply call `forward()`
        and then run the post-processing as described in this method.

        Parameters
        ----------
        images : Any
            See `preprocess_input()` for the allowed input images.

        Returns
        -------
        A tuple of `n` lists, where `n` is the number of input images.
        Each of the `n` lists will contain three values consisting of
        the bounding boxes, class labels, and prediction confidences
        for the corresponding input image.
        """
        # Process the images and run inference.
        images, shapes = self.preprocess_input(images, return_shapes = True)
        out = self._to_out(self.forward(images))

        # Post-process the output detections.
        boxes, confidences, labels = self._process_detections(out)
        boxes = self._rescale_bboxes(boxes, shapes)
        ret = self._remap_outputs(boxes, labels, confidences)
        return ret[0] if len(ret) == 1 else ret

    def show_prediction(self, image):
        """Shows the output predictions for one input image.

        This method is useful for instantly visualizing the predictions
        for a single input image. It accepts a single input image (or
        any type of valid 'image' input, as described in the method
        `preprocess_input()`), and then runs inference on that input
        image and displays its predictions in a matplotlib window.

        Parameters
        ----------
        image : Any
            See `preprocess_input()` for allowed types of inputs.

        Returns
        -------
        The matplotlib figure containing the image.
        """
        image = self._expand_input_images(image)[0]
        bboxes, labels, _ = self.predict(image)
        if isinstance(labels, int):
            bboxes, labels = [bboxes], [labels]
        return visualize_image_and_boxes(image, bboxes, labels)

    def load_benchmark(self, dataset, strict = False):
        """Loads a benchmark for the given semantic segmentation dataset.

        This method is used to load pretrained weights for a specific AgML dataset.
        In essence, it serves as a wrapper for `load_state_dict`, directly getting
        the model from its save path in the AWS storage bucket. You can then use the
        `benchmark` property to access the metric value of the benchmark, as well as
        additional training parameters which you can use to train your own models.

        Parameters
        ----------
        dataset : str
            The name of the object detection benchmark to load.
        strict : bool
            Whether to require the same number of classes.

        Notes
        -----
        If the given benchmark has a different number of classes than this input model,
        then the class network will be loaded with random weights, while the remaining
        network weights (backbone, box network, etc.) will use the pretrained weights
        for the benchmark. This can be disabled by setting `strict = True`.
        """
        if source(dataset).tasks.ml != 'object_detection':
            raise ValueError(
                f"You are trying to load a benchmark for a "
                f"{source(dataset).tasks.ml} task ({dataset}) "
                f"in an object detection model.")

        # Check loading strictness.
        cs = source(dataset).num_classes == self._num_classes
        if strict:
            if not cs:
                raise ValueError(
                    f"You cannot load a benchmark for a dataset '{dataset}' "
                    f"with {source(dataset).num_classes} classes, while your "
                    f"model has {self._num_classes} classes. If you want to, "
                    f"then you need to set `strict = False`.")

        # Load the benchmark.
        state = self._get_benchmark(dataset)
        if not strict and not cs:
            self.reset_class_net(source(dataset).num_classes)
            log(f"Loading a state dict for {dataset} with "
                f"{source(dataset).num_classes}, while your "
                f"model has {self._num_classes} classes. The "
                f"class network will use random weights.")
        self.load_state_dict(state)
        if not strict and not cs:
            self.reset_class_net(self._num_classes)
        self._benchmark = BenchmarkMetadata(dataset)

    def evaluate(self, loader, iou_threshold = 0.5, method = 'accumulate'):
        """Runs a mean average precision evaluation on the given loader.

        This method will loop over the provided `AgMLDataLoader` and compute
        the mean average precision at the provided `iou_threshold`. This can
        be done using two methods. Using the method `average` will compute
        the mean average precision for each individual sample and then average
        over all of the samples, while using the method `accumulate` will
        compute the mean average precision over the entire dataset.

        Parameters
        ----------
        loader : AgMLDataLoader
            An object detection loader with the dataset you want to evaluate.
        iou_threshold : float
            The IoU threshold between a ground truth and predicted bounding
            box at which point they are considered the same.
        method : str
            The method to use, as described above.

        Returns
        -------
        The final calculated mean average precision.
        """
        if not 0 < iou_threshold < 1:
            raise ValueError(
                f"The `iou_threshold` must be between 0 and 1, got {iou_threshold}.")
        if method not in ['accumulate', 'average']:
            raise ValueError(
                f"Method must be either `accumulate` or `average`, got {method}.")

        # Construct the mean average precision accumulator and run the calculations.
        mean_ap = MeanAveragePrecision(
            num_classes = self._num_classes, iou_threshold = iou_threshold)
        bar = tqdm(loader, desc = "Calculating Mean Average Precision")
        if method == 'average':
            cumulative_maps = []
        for sample in bar:
            image, truth = sample
            true_box, true_label = truth['bbox'], truth['category_id']
            bboxes, labels, conf = self.predict(image)
            mean_ap.update(*(
                dict(boxes = bboxes, labels = labels, scores = conf),
                dict(boxes = true_box, labels = true_label)))

            # If averaging, then calculate and reset the mAP, otherwise continue.
            if method == 'average':
                res = mean_ap.compute()
                cumulative_maps.append(res) # noqa
                bar.set_postfix({'map': float(res)})
                mean_ap.reset()

        # Compute the final mAP.
        if method == 'average':
            result = sum(cumulative_maps) / len(cumulative_maps)
        else:
            result = mean_ap.compute()
        return result









# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import matplotlib.pyplot as plt

from agml.viz.tools import auto_resolve_image, show_when_allowed, format_image
from agml.backend.tftorch import as_scalar, is_array_like


def _inference_best_shape(n_images):
    """Inferences the best matplotlib row/column layout.

    This method searches for the two closest factors of the number
    `n_images`, and returns this tuple as the best shape, since this
    is the closest to a square grid as possible.
    """
    a, b, i = 1, n_images, 0
    while a < b:
        i += 1
        if n_images % i == 0:
            a = i
            b = n_images // a
    return [b, a]


@show_when_allowed
@auto_resolve_image
def visualize_images(images, shape = None):
    """Visualizes a set of images in the given shape.

    Given a set of images, this method will generate a grid for the
    images and display them as such. The shape of the grid will by
    default be inferenced to be the two closest factors of the number
    of images (to  be as close to square as possible).

    Parameters
    ----------
    images : Any
        Either a list of images, a tuple of images and labels, or a list
        of image/label pairs (like you would get as the output of a dataset).
    shape : Any
        The shape of the display grid.

    Returns
    -------
    The matplotlib figure with the plotted images.
    """
    # If a prime number is passed, e.g. 23, then the `_inference_best_shape`
    # method will return the shape of (23, 1). Likely, the user is expecting
    # a non-rectangular shape such as (6, 4), where the bottom right axis is
    # empty. This method does not support such computations (yet).
    if shape is None:
        shape = _inference_best_shape(len(images))
    if max(shape) > 20:
        raise NotImplementedError(
            "Length of maximum shape length is greater than 20. "
            "This method does not support non-rectangular shapes.")

    fig, axes = plt.subplots(
        shape[0], shape[1], figsize = (shape[1] * 2, shape[0] * 2))
    try:
        iter_ax = axes.flat
    except AttributeError: # If showing only a single image.
        iter_ax = [axes]
    for image, ax in zip(images, iter_ax):
        ax.imshow(format_image(image))
        ax.set_aspect(1)
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.tick_params(
            axis = 'both', which = 'both', bottom = False,
            top = False, left = False, right = False
        )
        plt.setp(ax.spines.values(), visible = False)

    fig.tight_layout()
    return fig



@show_when_allowed
@auto_resolve_image
def visualize_images_with_labels(images, labels = None, *, info = None, shape = None):
    """Visualizes a set of images with their classification labels.

    Given a set of images and their corresponding labels, this method
    will generate a grid for the images and display them with their
    image classification labels displayed underneath them. The shape of
    the grid will by default be inferenced to be the two closest factors
    of the number of images (to be as close to square as possible).

    If you provide an `info` parameter, which will consist of the `info`
    property of an AgMLDataLoader (literally pass `loader.info`), then the
    method will convert the classification numbers to their label names.

    Parameters
    ----------
    images : Any
        Either a list of images, a tuple of images and labels, or a list
        of image/label pairs (like you would get as the output of a dataset).
    labels : Any
        A list or array of classification labels.
    info : DatasetMetadata
        The `loader.info` attribute of a dataloader.
    shape : Any
        The shape of the display grid.

    Returns
    -------
    The matplotlib figure with the plotted info.
    """
    if images is not None and labels is None:
        if is_array_like(images[0]):
            if images[0].ndim >= 3:
                images, labels = images[0], images[1]
            else:
                raise ValueError(
                    "If passing a numpy array for images, expected at "
                    "least three dimensions: (batch, height, width).")
        elif isinstance(images[0], (tuple, list)):
            if isinstance(images[0][0], np.ndarray):
                if len(images[0]) == 2:
                    _images, _labels = [], []
                    for content in images:
                        _images.append(content[0])
                        _labels.append(content[1])
                    images, labels = _images, _labels
                else:
                    images, labels = images[0], images[1]
    if labels is None:
        raise TypeError(
            "Invalid format for `images` and `labels`, see documentation.")
    if isinstance(images, np.ndarray) and images.shape[0] > 100:
        images, labels = [images], [labels]

    # Check if the labels are converted to one-hot, and re-convert them back.
    if is_array_like(labels):
        if labels.ndim == 2: # noqa
            labels = np.argmax(labels, axis = -1)

    # If a prime number is passed, e.g. 23, then the `_inference_best_shape`
    # method will return the shape of (23, 1). Likely, the user is expecting
    # a non-rectangular shape such as (6, 4), where the bottom right axis is
    # empty. This method does not support such computations (yet).
    if shape is None:
        shape = _inference_best_shape(len(images))
    if max(shape) > 20:
        raise NotImplementedError(
            "Length of maximum shape length is greater than 20. "
            "This method does not support non-rectangular shapes.")

    fig, axes = plt.subplots(
        shape[0], shape[1], figsize = (shape[1] * 2, shape[0] * 2))
    try:
        iter_ax = axes.flat
    except AttributeError: # If showing only a single image.
        iter_ax = [axes]
    for image, label, ax in zip(images, labels, iter_ax):
        ax.imshow(format_image(image))
        ax.set_aspect(1)
        label = as_scalar(label)
        if info is not None:
            label = info.num_to_class[label]
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.tick_params(
            axis = 'both', which = 'both', bottom = False,
            top = False, left = False, right = False
        )
        plt.setp(ax.spines.values(), visible = False)
        ax.set_xlabel(label)

    fig.tight_layout()
    return fig






# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

import numpy as np

from agml.utils.downloads import download_dataset as _download # noqa
from agml.utils.data import load_public_sources
from agml.data.metadata import DatasetMetadata
from agml.backend.config import data_save_path


class _PublicSourceFilter(object):
    """Filters public datasets based on the input filters."""
    _sources = load_public_sources()
    _current_filtered_source = []

    def apply_filters(self, **filters):
        if len(filters) == 0:
            self._current_filtered_source = self._sources.keys()
            return self
        source_sets = []
        for key, value in filters.items():
            if key not in list(self._sources.values())[0].keys():
                raise ValueError(f"Invalid filter: {key}.")
            internal_set = []
            if key == 'location':
                self._location_case(value, internal_set)
            elif key in ['n_images', 'num_images'] and value.startswith('>'):
                self._n_image_case_greater(
                    int(float(value[1:])), internal_set)
            elif key in ['n_images', 'num_images'] and value.startswith('<'):
                self._n_image_case_greater(
                    int(float(value[1:])), internal_set)
            else:
                for source_ in self._sources.keys():
                    try:
                        if self._sources[source_][key] == value:
                            internal_set.append(source_)
                    except: # some situations don't have certain arguments.
                        continue
            source_sets.append(internal_set)
        self._current_filtered_source \
            = functools.reduce(np.intersect1d, source_sets)
        return self

    def _location_case(self, desired, value_set):
        for source_ in self._sources.keys():
            param, value = desired.split(':')
            try:
                if self._sources[source_]['location'][param] == value:
                    value_set.append(source_)
            except KeyError: # some situations don't have location.
                continue
        return value_set

    def _n_image_case_greater(self, thresh, value_set):
        for source_ in self._sources.keys():
            try:
                if int(float(self._sources[source_]['n_images'])) >= thresh:
                    value_set.append(source_)
            except KeyError: # some situations don't have n_images.
                continue
        return

    def _n_image_case_lesser(self, thresh, value_set):
        for source_ in self._sources.keys():
            try:
                if int(float(self._sources[source_]['n_images'])) <= thresh:
                    value_set.append(source_)
            except KeyError: # some situations don't have n_images.
                continue
        return

    def print_result(self):
        return "[%s]" % ', '.join(self._current_filtered_source)

    def result(self):
        return [DatasetMetadata(s) for s in self._current_filtered_source]


def public_data_sources(**filters):
    """Lists all of the public data sources in AgML.

    The `filters` argument can be used to get a list of datasets
    which are constrained to a specific value, such as in the following:

    Get a list of all image classification datasets in AgML.
    > public_data_sources(ml_task = 'image_classification')

    The `location` and `n_images` arguments have a different procedure.
    For location, the argument needs to be passed as follows:

    > public_data_sources(location = 'continent:africa')
    > public_data_sources(location = 'country:denmark')

    For number of images, the argument can either be an exact integer,
    or if checking a threshold, then it should be passed as follows
    (note that the threshold is inclusive):

    > public_data_sources(n_images = 1000)
    > public_data_sources(n_images = ">1500")

    This method returns a list of `DatasetMetadata` objects, which can
    then be further inspected by additional conditions.

    Parameters
    ----------
    filters : Any
        An arbitrary number of keyword arguments representing different
        filters which should be applied to the list of data sources.

    Returns
    -------
    A list of sources filtered by the passed arguments.
    """
    s = _PublicSourceFilter().apply_filters(**filters)
    return s.result()


def source(name):
    """Returns the metadata of a single public dataset source.

    If you just want to inspect the info of one dataset, then you can use
    this method with `name` representing a dataset to get a single piece
    of `DatasetMetadata` which contains its information.

    Parameters
    ----------
    name : str
        The name of the dataset you want to inspect.

    Returns
    -------
    The metadata of the dataset `name`.
    """
    return DatasetMetadata(name)


def download_public_dataset(dataset, location = None, redownload = False):
    """Downloads a public dataset from AgML to a directory.

    While the `AgMLDataLoader` exists to load data directly into an
    accessor class, if you want to simply download the data to
    a location, whether for inspection or other purposes, this
    method enables such downloading. By default, data will be
    downloaded to the `~/.agml/datasets` directory.

    If you want to download multiple datasets at once, simply pass
    a sequence of datasets to the `dataset` argument.

    Parameters
    ----------
    dataset : str, list
        The name of the dataset(s) you want to download.
    location : str
        The local location you want to download the data to.
    redownload : bool
        Whether to download the dataset irrespective of existence.

    Returns
    -------
    The local directory of the dataset.
    """
    location = location if location \
                        else data_save_path()
    if not isinstance(dataset, (list, tuple, set, np.ndarray)):
        dataset = [dataset]
    for d in dataset:
        _download(d, location, redownload = redownload)
    return location





# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections

from agml.framework import AgMLSerializable
from agml.utils.data import load_model_benchmarks


# Named tuples which are used by the metadata.
Metric = collections.namedtuple('Metric', ['name', 'value'])


class BenchmarkMetadata(AgMLSerializable):
    """Contains metadata regarding a specific benchmark on a dataset.

    When loading a pretrained model corresponding to a certain benchmark,
    this class is used to provide information regarding the benchmark,
    as well as key hyperparameters which can be used for reproducibility.
    """
    serializable = frozenset(('dataset', 'meta'))

    def __init__(self, dataset):
        # Load the information for the given dataset.
        if dataset is None:
            self._dataset = None
            self._meta = {'hyperparameters': {
                'model_config': None, 'optimizer_config': None,
                'metric': {None: None}, 'epochs_trained': None}}
            return
        try:
            self._dataset = dataset
            self._meta = load_model_benchmarks()[dataset]
        except KeyError:
            raise ValueError(f"Could not find a valid benchmark for dataset ({dataset}).")

    def __str__(self):
        return f"<Benchmark {self._dataset}>({self._meta})"

    @property
    def model_config(self):
        """Returns a dictionary with the configuration for the model."""
        return self._meta['hyperparameters']['model_config']

    @property
    def optimizer_config(self):
        """Returns a dictionary with the configuration for the optimizer."""
        return self._meta['hyperparameters']['optimizer_config']

    @property
    def epochs_trained(self):
        """Returns the number of epochs the benchmark was trained for."""
        return self._meta['hyperparameters']['epochs']

    @property
    def metric(self):
        """Returns the value of the metric for the benchmark."""
        name, value = list(self._meta['metric'].items())[0]
        return Metric(name = name, value = value)






# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Experimental features in AgML, these may be added or removed permanently."""

from agml.framework import AgMLSerializable


__all__ = ['allow_nested_data_splitting']


class AgMLExperimentalFeatureWrapper(AgMLSerializable):
    """Stores all of the feature values."""
    _NESTED_SPLITTING = False

    def __init__(self):
        raise ValueError("This class should not be instantiated!")

    @classmethod
    def allow_nested_data_splitting(cls, value: bool) -> None:
        """Enables/disables nested splitting of `AgMLDataLoader`s.

        This method can be used to either enable or disable a feature which
        allows an sub-AgMLDataLoader which has been split to a custom data
        split to be split again. This is an experimental feature, which
        would allow for multiple levels of nested splits (e.g., you could
        have a `loader.train_data.val_data.test_data`).

        Parameters
        ----------
        value : bool
            Whether to enable or disable the feature.

        Notes
        -----
        This method must be called at the start of each script in order
        to function, otherwise it will default to `False`.
        """
        if not isinstance(value, bool):
            raise TypeError("Expected either True or False.")
        cls._NESTED_SPLITTING = value

    @classmethod
    def nested_splitting(cls):
        return cls._NESTED_SPLITTING


# While the `AgMLExperimentalFeatureWrapper` class controls the actual values,
# we expose each of its toggle methods as part of the `agml.backend` API.
allow_nested_data_splitting = AgMLExperimentalFeatureWrapper.allow_nested_data_splitting




# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

import torch
import torch.nn as nn

from agml.utils.logging import log


@torch.jit.script
def bbox_iou(predicted_box, truth_box):
    """Calculates the IOU of predicted and truth bounding boxes."""
    # Calculate the coordinates of the bounding box overlap.
    box1_x1 = predicted_box[..., 0:1]
    box1_y1 = predicted_box[..., 1:2]
    box1_x2 = predicted_box[..., 0:1] + predicted_box[..., 2:3]
    box1_y2 = predicted_box[..., 1:2] + predicted_box[..., 3:4]
    box2_x1 = truth_box[..., 0:1]
    box2_y1 = truth_box[..., 1:2]
    box2_x2 = truth_box[..., 0:1] + truth_box[..., 2:3]
    box2_y2 = truth_box[..., 1:2] + truth_box[..., 3:4]
    x1 = torch.max(box1_x1, box2_x1)
    y1 = torch.max(box1_y1, box2_y1)
    x2 = torch.min(box1_x2, box2_x2)
    y2 = torch.min(box1_y2, box2_y2)

    # Get the area of the union.
    intersection = (x2 - x1).clamp(0) * (y2 - y1).clamp(0)
    box1_area = abs((box1_x2 - box1_x1) * (box1_y2 - box1_y1))
    box2_area = abs((box2_x2 - box2_x1) * (box2_y2 - box2_y1))
    union_area = (box1_area + box2_area - intersection)

    # Return the intersection over union.
    return intersection / (union_area + 1e-6)


@torch.no_grad()
def mean_average_precision(
        predicted_boxes, truth_boxes, num_classes = 1, iou_thresh = 0.5):
    """Calculates the mean average precision for predicted and true boxes."""
    average_precisions = []

    # Check whether to add confidence scores to the ground truth boxes.
    if len(truth_boxes[0]) == 6:
        out_boxes = []
        for a in truth_boxes:
            out_boxes.append([*a[:2], 1.0, *a[2:]])
        truth_boxes = out_boxes.copy()

    # Calculate average precision for each class.
    pred_boxes, true_boxes = torch.tensor(predicted_boxes), torch.tensor(truth_boxes)
    for c in range(num_classes):
        # Get the predictions and targets corresponding to this class.
        detections = pred_boxes[torch.where(pred_boxes[:, 1] == c)[0]].tolist()
        ground_truths = true_boxes[torch.where(true_boxes[:, 1] == c)[0]].tolist()

        # If there are no ground truths, then the per-class AP is 0.
        if len(ground_truths) == 0:
            average_precisions.append(0.0)
            continue

        # Get all of the unique data samples and create a dictionary
        # storing all of the corresponding bounding boxes for each sample.
        training_ids = torch.unique(true_boxes[:, 0])
        truth_samples_by_id = {
            idx.numpy().item(): true_boxes[torch.where(true_boxes[:, 0] == idx)] # noqa
            for idx in training_ids}

        # Determine the number of boxes for each of the training samples.
        numpy_gt = torch.tensor(ground_truths)
        amount_bboxes = {int(k.numpy().item()): torch.zeros(v) for k, v in zip(
            *torch.unique(numpy_gt[:, 0], return_counts = True))}

        # Sort the boxes by probabilities.
        detections.sort(key = lambda x: x[2], reverse = True)
        tp = torch.zeros((len(detections)))
        fp = torch.zeros((len(detections)))
        total_true_bboxes = len(ground_truths)

        # If there are no boxes for this class, then there are
        # no calculations to do, so skip it.
        if total_true_bboxes == 0:
            continue

        for detection_idx, detection in enumerate(detections):
            # Get the update number for this detection.
            update_num = int(detection[0])

            # Only take out the ground_truths that have the same
            # training idx as the detection.
            ground_truth_img = truth_samples_by_id[update_num]

            # Get the bounding box with the highest IoU.
            ious = torch.tensor([bbox_iou(
                torch.tensor(detection[3:]), gt[3:].clone())
                for gt in ground_truth_img])
            best_iou = torch.max(ious)
            best_gt_idx = torch.argmax(ious)

            # If the IoU is above the threshold, then it may be a true positive.'
            if best_iou > iou_thresh:
                # This should be the first time the box is detected. Otherwise,
                # that would mean that there are multiple predicted bounding
                # boxes for the same object, which is a false positive.
                try:
                    if amount_bboxes[update_num][best_gt_idx] == 0:
                        tp[detection_idx] = 1
                        amount_bboxes[update_num][best_gt_idx] = 1
                    else:
                        fp[detection_idx] = 1
                except KeyError:
                    # A false detection.
                    fp[detection_idx] = 1

            # If the IoU is below the threshold, then it is a false positive.
            else:
                fp[detection_idx] = 1

        # Calculate the prediction/recalls and update the array.
        tp_cumsum = torch.cumsum(tp, dim = 0)
        recalls = tp_cumsum / (total_true_bboxes + 1e-6)
        precisions = tp_cumsum / (tp_cumsum + torch.cumsum(fp, dim = 0) + 1e-6)
        precisions = torch.cat((torch.tensor([1]), precisions))
        recalls = torch.cat((torch.tensor([0]), recalls))
        average_precisions.append(torch.trapz(precisions, recalls))

    # Calculate the mean of all of the average precisions.
    return sum(average_precisions) / len(average_precisions)


class MeanAveragePrecision(nn.Module):
    """A metric to calculate mean average precision for object detection tasks.

    This class is used as a wrapper around mean average precision calculations,
    which allows for accumulation of predictions over time. The `update` (and
    `batch_update`) methods can be used to update data in the class, then `compute`
    to get the calculated mean average precision, and finally `reset` to restore
    the accumulators to an empty state, allowing from-scratch calculations.
    """

    def __init__(self, num_classes = 1, iou_threshold = 0.5):
        # Set base parameters.
        super(MeanAveragePrecision, self).__init__()
        self._num_classes = num_classes
        self._iou_threshold = iou_threshold

        # Store the truth and prediction data in containers.
        self._prediction_data, self._truth_data = [], []
        self._num_updates = 0

    @staticmethod
    def _scalar_to_array(*args):
        """Converts 0-dimensional scalar arrays to 1-d arrays."""
        cvt = lambda x: np.expand_dims(x, 0) if x.ndim == 0 else x
        outs = [cvt(arg) for arg in args]
        return outs[0] if len(args) == 1 else outs

    def update(self, pred_data, gt_data):
        """Update the tracker with prediction and ground truth data.

        The arguments `pred_data` and `gt_data` should be either dictionaries
        (with the following keys), or lists of values which correspond in order
        to the same keys listed below:

        - `pred_data`: `boxes`, `labels`, and `scores`.
        - `gt_data`: `boxes` and `labels`.

        Note: To update a batch of data, use `batch_update()`.
        """
        # Get the relevant data from the input arguments.
        if isinstance(pred_data, dict):
            pred_boxes, pred_labels, pred_scores = \
                pred_data['boxes'], pred_data['labels'], pred_data['scores']
        else:
            pred_boxes, pred_labels, pred_scores = pred_data
        if isinstance(gt_data, dict):
            gt_boxes, gt_labels = \
                gt_data['boxes'], gt_data['labels']
        else:
            gt_boxes, gt_labels = gt_data

        # Format the data.
        pred_boxes = np.squeeze(pred_boxes)
        pred_labels = np.squeeze(pred_labels)
        pred_scores = np.squeeze(pred_scores)
        pred_labels, gt_labels, pred_scores = \
            self._scalar_to_array(pred_labels, gt_labels, pred_scores)
        if pred_boxes.ndim == 1:
            pred_boxes = np.expand_dims(pred_boxes, axis = 0)
        if gt_boxes.ndim == 1:
            gt_boxes = np.expand_dims(gt_boxes, axis = 0)

        # Create the data in the proper format.
        for bbox, label in zip(gt_boxes, gt_labels):
            self._truth_data.append(
                [self._num_updates, int(label - 1), 1.0, *bbox])
        if pred_boxes.ndim == 1:
            pred_boxes = np.expand_dims(pred_boxes, axis = 0)
        for bbox, label, score in zip(pred_boxes, pred_labels, pred_scores):
            self._prediction_data.append(
                [self._num_updates, int(label - 1), score, *bbox])

        # Increment the number of updates.
        self._num_updates += 1

    def compute(self):
        """Computes the mean average precision with the given data."""
        if self._num_updates == 0:
            log("Tried to compute mean average precision "
                "without any data updates; returning 0.0.")
            return 0.0

        return mean_average_precision(
            predicted_boxes = self._prediction_data,
            truth_boxes = self._truth_data,
            num_classes = self._num_classes,
            iou_thresh = self._iou_threshold,
        )

    def reset(self):
        """Resets the mean average precision."""
        del self._prediction_data, self._truth_data
        self._prediction_data, self._truth_data = [], []
        self._num_updates = 0








# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from agml.backend.tftorch import is_array_like
from agml.utils.general import resolve_list_value


def _resolve_coco_annotations(annotations):
    """Resolves COCO annotations to a standard format.

    Generally, COCO annotations will either be represented as a
    dictionary which contains multiple elements for each of its keys,
    e.g. multiple bounding boxes and areas, or a list of multiple COCO
    dictionaries. This method resolves it into the former case.
    """
    if is_array_like(annotations) and not isinstance(annotations, np.ndarray):
        return annotations.numpy()
    if isinstance(annotations, np.ndarray):
        return annotations
    if isinstance(annotations, list):
        if not isinstance(annotations[0], dict):
            return annotations
        if len(annotations) == 1:
            return annotations
        annotation = {'bboxes': [], 'labels': [], 'area': [],
                      'image_id': "", 'iscrowd': [], 'segmentation': []}
        for a_set in annotations:
            annotation['bboxes'].append(a_set['bbox'])
            annotation['labels'].append(a_set['category_id'])
            annotation['iscrowd'].append(a_set['iscrowd'])
            annotation['segmentation'].append(a_set['segmentation'])
            annotation['area'].append(a_set['area'])
            annotation['image_id'] = a_set['image_id']
        for key, value in annotation.items():
            out = np.array(value)
            if np.isscalar(out):
                out = out.item()
            annotation[key] = out
        return annotation
    elif isinstance(annotations, dict):
        return annotations
    else:
        raise TypeError(
            "Expected either a single COCO annotation "
            "dictionary or a list of multiple dictionaries.")


def coco_to_bboxes(annotations):
    """Extracts the bounding boxes and labels from COCO JSON annotations.

    Given either a list of COCO JSON annotations, or a single
    dictionary with multiple values, this method will extract the
    bounding boxes and category labels from the annotations and
    return just the bounding boxes and category labels in two arrays.

    Parameters
    ----------
    annotations : {list, dict}
        The COCO JSON annotations in list/dict format.

    Returns
    -------
    Two arrays consisting of the bounding boxes and labels.
    """
    annotations = _resolve_coco_annotations(annotations)
    return annotations['bboxes'], annotations['labels']


def convert_bbox_format(annotations_or_bboxes, fmt):
    """Converts bounding box formats for COCO JSON and others.

    This method converts the format of bounding boxes as specified
    in the 'fmt' argument, which describes the format of the bounding
    boxes passed to the 'annotations_or_bboxes' argument. From there,
    it will convert to the standard COCO JSON bounding box format,
    namely ('x1', 'y1' 'width', 'height'). This method supports the
    following conversions (note that 'x1' and 'y1' are the top-left
    coordinates, and 'x2' and 'y2' are top-right):

    1. ('x1', 'x2', 'y1', 'y2') to COCO JSON.
    2. ('x_min', 'y_min', 'x_max', 'y_max') to COCO JSON. This format
        can also be passed with the simple string `pascal_voc`.
    3. ('x_min', 'y_min', 'width', 'height') to COCO JSON.

    Note that the variables in the bounding boxes in the above format
    can be in any order, this just needs to be reflected in the 'fmt'
    argument, and it should contain some combination of the above.

    Parameters
    ----------
    annotations_or_bboxes : {np.ndarray, list, dict}
        Either a COCO JSON annotation dictionary, or a numpy array/list
        with all of the annotations.
    fmt : {list, tuple}
        A list or tuple with one of the above formats.

    Returns
    -------
    The initial argument type (either dict or array) with the bounding
    boxes formatted in the COCO JSON format.
    """
    annotations_or_bboxes = _resolve_coco_annotations(annotations_or_bboxes)
    if isinstance(annotations_or_bboxes, dict):
        annotations = annotations_or_bboxes['bboxes']
    else:
        annotations = annotations_or_bboxes
        if isinstance(annotations[0], (int, float)):
            annotations = [annotations]
    if isinstance(fmt, str):
        if 'voc' in fmt or 'pascal' in fmt:
            fmt = 'x_min y_min x_max y_max'
        elif 'efficientdet' in fmt or 'effdet' in fmt:
            fmt = 'y_min x_min y_max x_max'
        if ',' in fmt:
            fmt = fmt.split(',')
        else:
            fmt = fmt.split(' ')
    if len(fmt) != 4:
        raise ValueError(f"Argument 'fmt' should contain 4 values, got {len(fmt)}.")
    
    # Define all of the intermediate conversion methods
    def _x1_x2_y1_y2_to_coco(annotation): # noqa
        x1, x2, y1, y2 = annotation
        width, height = abs(x2 - x1), abs(y2 - y1)
        return [x1, y1, width, height]
    def _xmin_ymin_xmax_ymax_to_coco(annotation): # noqa
        xmin, ymin, xmax, ymax = annotation
        width, height = abs(xmax - xmin), abs(ymax - ymin)
        return [xmin, ymin, width, height]
    def _xmin_ymin_width_height_to_coco(annotation): # noqa
        xmin, ymin, width, height = annotation
        x1, y1 = xmin, ymin - height
        return [x1, y1, width, height]
    def _x1_y1_width_height_to_coco(annotation): # noqa
        return annotation # This is just here for reordering.

    # Resolve the format
    fmt_bases = [['x1', 'x2', 'y1', 'y2'],
                 ['x_min', 'y_min', 'x_max', 'y_max'],
                 ['x_min', 'y_min', 'width', 'height'],
                 ['x1', 'y1', 'width', 'height']]
    fmt_map = {0: _x1_x2_y1_y2_to_coco,
               1: _xmin_ymin_xmax_ymax_to_coco,
               2: _xmin_ymin_width_height_to_coco,
               3: _x1_y1_width_height_to_coco}
    fmt_found = False
    map_fmt, select_order = None, None
    for indx, base in enumerate(fmt_bases):
        if all(i in base for i in fmt):
            fmt_found, map_fmt = True, fmt_map[indx]
            select_order = [base.index(i) for i in fmt]
    if not fmt_found:
        raise ValueError(
            f"Invalid format {fmt}, see `convert_bbox_format` "
            f"for information about valid formats.")

    # Convert the formats
    formatted_annotations = []
    for bbox in annotations:
        sorted_bbox = np.array(bbox)[select_order]
        formatted_annotations.append(map_fmt(sorted_bbox))
    formatted_annotations = np.array(formatted_annotations)
    if isinstance(annotations_or_bboxes, dict):
        res = annotations_or_bboxes.copy()
        res['bboxes'] = formatted_annotations
        return res
    return resolve_list_value(formatted_annotations)




# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from agml.framework import AgMLSerializable
from agml.data.object import DataObject
from agml.data.builder import DataBuilder
from agml.data.metadata import DatasetMetadata
from agml.data.managers.transforms import TransformManager
from agml.data.managers.resize import ImageResizeManager
from agml.data.managers.training import TrainingManager

from agml.utils.general import NoArgument
from agml.backend.tftorch import convert_to_batch, is_array_like
from agml.utils.random import seed_context
from agml.utils.image import consistent_shapes
from agml.utils.logging import log


class DataManager(AgMLSerializable):
    """Manages the data for a dataset loaded in an `AgMLDataLoader`.

    The `DataManager` is the core internal object which, as its name
    suggests, manages the data being used in an `AgMLDataLoader`. Upon
    instantiation of an `AgMLDataLoader`, the internal data, which is
    represented generally as image paths and corresponding annotations,
    is placed into a `DataManager` container.

    Accessing data from an `AgMLDataLoader` calls the `DataManager`,
    which takes into account all of the potential splits, transforms,
    and batching that may have been applied to the loader.

    This enables more streamlined processing and transforming of data,
    as well as parallelizing data loading as desired (in the future).
    Another way to think about the `DataManager` is it being a list
    of `DataObjects`, with extra logic for processing.
    """
    serializable = frozenset((
        'data_objects', 'resize_manager', 'accessors', 'task',
        'dataset_name', 'shuffle', 'batch_size', 'dataset_root',
        'transform_manager', 'builder', 'train_manager'))

    def __init__(self, builder, task, name, root, **kwargs):
        # Set the basic class information.
        self._task = task
        self._dataset_name = name
        self._dataset_root = root

        # Create the `DataObject`s from the `DataBuilder`.
        if not isinstance(builder, DataBuilder):
            builder = DataBuilder.from_data(
                contents = builder,
                info = DatasetMetadata(name),
                root = root)
        self._builder = builder
        self._create_objects(self._builder, task)

        # Set up the internal transform managers. These control
        # the application of transformations to the input data.
        self._transform_manager = TransformManager(task = task)
        self._resize_manager = ImageResizeManager(
            task = task, dataset = name, root = root
        )

        # The transform and resize managers are wrapped inside of a
        # `TrainingManager`, which controls the application of
        # preprocessing to the data based on the class state.
        self._train_manager = TrainingManager(
            transform_manager = self._transform_manager,
            resize_manager = self._resize_manager,
            task = task
        )

        # While we store data in the list of `DataObject`s, the actual
        # accessing of the data by index doesn't happen by directly
        # accessing that list, but instead by accessing the array of
        # indices below. This is useful for two reasons.
        #
        # It allows a much more simpler way to store and access the
        # state of the data. E.g., when splitting the data or accessing
        # it, we only need to check through this array rather than
        # searching through or shuffling the actual list of objects.
        #
        # Secondly, it makes batching data much more straightforward and
        # allows for storing representations of the data in different
        # formats, e.g., shuffling, to be done without having to interfere
        # with the actual `DataObject`s, which is much more convenient.
        self._accessors = np.arange(len(self._data_objects))

        # The following parameters store various parameters which are
        # used internally or accessed by the `AgMLDataLoader` externally.
        self._batch_size = None
        self._shuffle = kwargs.get('shuffle', True)
        self._maybe_shuffle()

    def data_length(self):
        """Calculates the length of the data based on the batching state."""
        return len(self._accessors)

    def _create_objects(self, builder, task):
        """Creates `DataObject`s from the provided content.

        Here, `content` is a dictionary mapping an an input data piece,
        an image, with its corresponding expected output, its annotation.
        """
        self._data_objects = []
        contents = builder.get_contents()
        if 'inputs' in contents.keys():
            contents = zip(tuple(contents['inputs']),
                           tuple(contents['outputs']))
        else:
            contents = contents.items()
        for content in list(contents):
            self._data_objects.append(DataObject.create(
                contents = content, task = task,
                root = self._dataset_root))

    def _maybe_shuffle(self, seed = None):
        """Wraps automatic shuffling to see if it is enabled or not."""
        if self._shuffle:
            self.shuffle(seed = seed)

    def update_train_state(self, state):
        """Updates the training state in the `TrainingManager`."""
        self._train_manager.update_state(state)

    def shuffle(self, seed = None):
        """Shuffles the contents of the `DataManager`.

        This method simply shuffles the order of the `DataObject`s
        which are stored inside this `DataManager`. Optionally, a seed
        can be provided to shuffle them inside of a specific context.
        """
        if seed is None:
            np.random.shuffle(self._accessors)
        else:
            with seed_context(seed):
                np.random.shuffle(self._accessors)

    def generate_split_contents(self, splits):
        """Generates split contents given a dictionary of the split indexes.

        This method, given a set of data split indexes, applies the indexing
        to the original content and gets a mapping of images and annotations
        which are returned back to the `AgMLDataLoader` to be constructed into
        `DataBuilder`s and wrapped into new `DataManager`s.
        """
        if self._task == 'object_detection':
            contents = np.array(list(
                self._builder.get_contents().items()), dtype = object)
        else:
            contents = np.array(list(self._builder.get_contents().items()))
        try:
            return {k: dict(contents[v]) for k, v in splits.items()}
        except IndexError:
            raise Exception(
                f"Could not generate split contents, likely due to an error "
                f"with the metadata for the dataset `{self._dataset_name}`. "
                f"Please raise this error with the AgML team.")

    def batch_data(self, batch_size):
        """Batches the data into consistent groups.

        The batched data is stored inside of this manager, as a set of
        indexes which are read and loaded when the data is accessed.
        See the information above the `_accessors` parameter above.
        """
        # If the data is already batched and a new batch size is called,
        # then update the existing batch sizes. For unbatching the data,
        # update the batch state and then flatten the accessor array.
        if self._batch_size is not None:
            try:
                self._accessors = np.concatenate(self._accessors).ravel()
            except ValueError:
                # The array is currently 0-dimensional.
                pass
        if batch_size is None or batch_size == 0:
            self._batch_size = None
            return

        # If we have a batch size of `1`, then don't do anything
        # since this doesn't really mean to do anything.
        if batch_size == 1:
            return

        # Otherwise, calculate the actual batches and the overflow
        # of the contents, and then update the accessor.
        num_splits = len(self._accessors) // batch_size
        data_items = np.array(self._accessors)
        overflow = len(self._accessors) - num_splits * batch_size
        extra_items = data_items[-overflow:]
        try:
            batches = np.array_split(
                np.array(self._accessors
                         [:num_splits * batch_size]), num_splits)
        except ValueError:
            log(f"There is less data ({len(self._accessors)}) than the provided "
                f"batch size ({batch_size}). Consider using a smaller batch size.")
            batches = [self._accessors]
        else:
            if len(extra_items) < batch_size:
                batches.append(extra_items)
        self._accessors = np.array(batches, dtype = object)
        self._batch_size = batch_size

    def assign_resize(self, image_size, method):
        """Assigns a resizing factor for the image and annotation data."""
        if image_size is None:
            image_size = 'default'
        self._resize_manager.assign(image_size, method)

    def push_transforms(self, **transform_dict):
        """Pushes a transformation to the data transform pipeline."""
        # Check if any transforms are being reset and assign them as such.
        if all(i is NoArgument for i in transform_dict):
            transform_dict = {
                'transform': 'reset',
                'target_transform': 'reset',
                'dual_transform': 'reset'
            }
        else:
            empty_keys, reset_keys = [], []
            for key, value in transform_dict.items():
                if value is NoArgument:
                    empty_keys.append(key)
                if value is None:
                    reset_keys.append(key)
            if len(empty_keys) != 0:
                for key, value in transform_dict.items():
                    if value is None:
                        transform_dict[key] = 'reset'
                    elif value is NoArgument:
                        transform_dict[key] = None

        # There is no `dual_transform` object for image classification.
        if self._task == 'image_classification':
            if transform_dict['dual_transform'] is None:
                transform_dict['dual_transform'] = 'reset'

        # Assign the transforms to the manager in order.
        for key in ['transform', 'target_transform', 'dual_transform']:
            self._transform_manager.assign(key, transform_dict[key])

    def _load_one_image_and_annotation(self, obj):
        """Loads one image and annotation from a `DataObject`."""
        return self._train_manager.apply(
            obj = obj, batch_state = self._batch_size is not None
        )

    def _load_multiple_items(self, indexes):
        """Loads multiple images and annotations from a set of `DataObject`s."""
        # Either we're getting multiple batches, or just multiple items.
        contents = []
        if self._batch_size is not None:
            for i in indexes:
                contents.append(self._load_batch(self._accessors[i]))
        else:
            for i in indexes:
                contents.append(self._load_one_image_and_annotation(
                    self._data_objects[self._accessors[i]]))
        return contents

    def _batch_multi_image_inputs(self, images):
        """Converts either a list of images or multiple input types into a batch."""
        # If the input images are just a simple batch.
        if is_array_like(images[0]):
            return convert_to_batch(images)

        # Otherwise, convert all of them independently.
        keys = images[0].keys()
        batches = {k: [] for k in keys}
        for sample in images:
            for key in sample:
                batches[key].append(sample[key])
        return {k: self._batch_multi_image_inputs(i) for k, i in batches.items()}

    def _batch_multi_output_annotations(self, annotations):
        """Converts either a list of annotations or multiple annotation types into a batch."""
        # If the output annotations are simple objects.
        if (isinstance(annotations[0], (list, np.ndarray))
            or isinstance(annotations, (list, np.ndarray))
                and isinstance(annotations[0], (int, float))):
            if not consistent_shapes(annotations):
                annotations = np.array(annotations, dtype = object)
            else:
                annotations = np.array(annotations)
            return annotations

        # For object detection, just return the COCO JSON dictionaries.
        if self._task == 'object_detection':
            return annotations

        # Otherwise, convert all of them independently.
        keys = annotations[0].keys()
        batches = {k: [] for k in keys}
        for sample in annotations:
            for key in sample:
                batches[key].append(sample[key])
        return {k: self._batch_multi_output_annotations(i) for k, i in batches.items()}

    def _load_batch(self, batch_indexes):
        """Gets a batch of data from the dataset.

        This differs from simply getting multiple pieces of data from the
        dataset, such as a slice, in that it also stacks the data together
        into a valid batch and returns it as such.
        """
        # Get the images and annotations from the data objects.
        images, annotations = [], []
        for index in batch_indexes:
            image, annotation = self._load_one_image_and_annotation(
                self._data_objects[index])
            images.append(image)
            annotations.append(annotation)

        # Attempt to create batched image arrays.
        images = self._batch_multi_image_inputs(images)

        # Attempt the same for the annotation arrays. This is more complex
        # since there are many different types of annotations, namely labels,
        # annotation masks, COCO JSON dictionaries, etc. We need to properly
        # create a batch in each of these cases.
        annotations = self._batch_multi_output_annotations(annotations)

        # Return the batches.
        return self._train_manager.make_batch(
            images = images,
            annotations = annotations
        )

    def get(self, indexes):
        """Loads and processes a piece (or pieces) of data from the dataset.

        This is the actual accessor method that performs the loading of data
        and the relevant processing as dictated by loading, image resizing,
        transform application, and other internal processing methods such as
        creating batches. This is called by the `AgMLDataLoader` to get data.
        """
        # If there is only one index and the data is not batched,
        # then we just need to return a single `DataObject`.
        if isinstance(indexes, int) and self._batch_size is None:
            return self._load_one_image_and_annotation(
                self._data_objects[self._accessors[indexes]]
            )

        # If we have a batch of images, then return the batch.
        if isinstance(indexes, int) and self._batch_size is not None:
            return self._load_batch(self._accessors[indexes])

        # Otherwise, if there are multiple indexes (e.g., an unstacked
        # slice or just a tuple of integers), then we get multiple images.
        if isinstance(indexes, (list, tuple)):
            return self._load_multiple_items(indexes)

    











# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import glob

import cv2

import matplotlib.pyplot as plt
from matplotlib.patches import FancyArrowPatch
from mpl_toolkits.mplot3d import proj3d

from agml.viz.tools import show_when_allowed
from agml.viz.labels import _inference_best_shape
from agml.synthetic.tools import _is_agml_converted
from agml.utils.image import imread_context


class Arrow3D(FancyArrowPatch):
    """Draws a 3-dimensional arrow using the given coordinates."""
    def __init__(self, xs, ys, zs, *args, **kwargs):
        FancyArrowPatch.__init__(self, (0, 0), (0, 0), *args, **kwargs)
        self._coords = xs, ys, zs

    def draw(self, renderer):
        xs, ys, zs = self._coords
        xs, ys, zs = proj3d.proj_transform(xs, ys, zs, self.axes.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
        super().draw(renderer)


@show_when_allowed
def plot_synthetic_camera_positions(positions, lookat):
    """Plots camera perspectives: positions and lookat vectors.

    This method is used to plot camera perspectives from a provided
    set of positional coordinates and lookat vector coordinates. This
    can be used to visualize the perspectives of cameras when generating
    synthetic agricultural data using Helios.

    The output of the method `agml.synthetic.generate_camera_positions`
    can be piped directly into this method to get a plot straight from
    environment parameters.

    Parameters
    ----------
    positions : list
        A list of 3-d coordinates indicating the positions of the
        cameras whose perspectives are to be plotted.
    lookat : list
        A list of lookat vector coordinates.

    Returns
    -------
    The matplotlib figure with the plot.
    """
    # Creat ethe figure.
    plt.figure(figsize = (7, 7))
    ax = plt.axes(projection = '3d')
    ax.plot3D(0, 0, 1, 'r*', label = 'Canopy Origin')
    for i in range(len(positions)):
        ax.plot3D(positions[i][0], positions[i][1], positions[i][2], 'o',
                  label = 'Camera ' + str(i))
        arw = Arrow3D([positions[i][0], lookat[i][0]],
                      [positions[i][1], lookat[i][1]],
                      [positions[i][2], lookat[i][2]],
                      arrowstyle = "->", color = "purple",
                      lw = 1, mutation_scale = 25)
        ax.add_artist(arw)
    ax.legend(loc = 'upper center', bbox_to_anchor = (0.5, 1.05),
              ncol = 4, fancybox = True, shadow = True)
    fig = plt.gcf()
    return fig


@show_when_allowed
def visualize_all_views(dataset_path, image):
    """Plots all of the camera views for a specific rendered canopy.

    Given a path to a set of camera views for a rendered canopy, this method
    will generate a grid onto which all of the views will be displayed. This
    enables inspection of a plant from multiple angles, supposing that there
    are multiple views generated for it.

    Note that the dataset can either be in the original Helios-generated format,
    or it can be converted. This method will either way find all of the views
    for the provided input image number.

    Parameters
    ----------
    dataset_path : str
        The path to the entire dataset.
    image : {int, str}
        An integer (possibly in string format) representing the number of the
        image whose camera views you want to view.

    Returns
    -------
    The matplotlib figure with the plot.
    """
    dataset_path = os.path.abspath(os.path.expanduser(dataset_path))
    if not os.path.exists(dataset_path):
        raise NotADirectoryError(
            f"The provided dataset path {dataset_path} does not exist.")

    # Get all of the views for the input image.
    image = str(image)
    if image.startswith('image'):
        image = image.replace('image', '')
    if _is_agml_converted(dataset_path):
        views = glob.glob(os.path.join(
            dataset_path, 'images', f'image{image}-view*.jpeg'))
        if len(views) == 0:
            views = glob.glob(os.path.join(
                dataset_path, f'image{image}-view*.jpeg'))
    else:
        views = glob.glob(os.path.join(
            dataset_path, f'image{image}', '**/*.jpeg'
        ), recursive = True)

    # Check that there are images.
    if len(views) == 0:
        raise FileNotFoundError(f"Could not find any views for image '{image}'.")

    # Parse all of the images.
    images = []
    for view in views:
        with imread_context(view) as img:
            images.append(cv2.cvtColor(img, cv2.COLOR_BGR2RGB))

    # Construct the figure.
    shape = _inference_best_shape(len(images))
    fig, axes = plt.subplots(shape[0], shape[1], figsize = (shape[1] * 5, shape[0] * 5))
    for image, ax in zip(images, axes.flat):
        ax.imshow(image)
        ax.set_axis_off()

    # Return the figure.
    fig.tight_layout()
    return fig






# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This part of the backend controls the AgML methods where either
TensorFlow or PyTorch methods can be used, and prevents unnecessary
importing of either library (which takes a significant amount of time).
"""
import types
import inspect
import logging
import importlib
import functools

import numpy as np

from agml.utils.logging import log
from agml.utils.image import consistent_shapes


# Suppress any irrelevant warnings which will pop up from either backend.
import warnings
warnings.filterwarnings(
    'ignore', category = UserWarning, message = '.*Named tensors.*Triggered internally.*')


class StrictBackendError(ValueError):
    def __init__(self, message = None, change = None, obj = None):
        if message is None:
            message = f"Backend was manually set to " \
                      f"'{get_backend()}', but got an object " \
                      f"from backend '{change}': {obj}."
        super(StrictBackendError, self).__init__(message)


# Check if TensorFlow and PyTorch exist in the environment.
_HAS_TENSORFLOW: bool
_HAS_TORCH: bool


@functools.lru_cache(maxsize = None)
def _check_tf_torch():
    global _HAS_TENSORFLOW, _HAS_TORCH
    try:
        import tensorflow
    except ImportError:
        _HAS_TENSORFlOW = False
    else:
        _HAS_TENSORFLOW = True
    try:
        import torch
    except ImportError:
        _HAS_TORCH = False
    else:
        _HAS_TORCH = True


# Default backend is PyTorch.
_BACKEND = 'torch'
_USER_SET_BACKEND = False


def get_backend():
    """Returns the current AgML backend."""
    return _BACKEND


def set_backend(backend):
    """Change the AgML backend for the current session.

    By default, AgML uses PyTorch as a backend, but it is
    compatible with both TensorFlow and PyTorch. AgML can
    also automatically inference the backend from the
    different parameters passed into `AgMLDataLoader` and
    other internal library methods.

    This method allows a user to automatically set the backend.
    """
    global _USER_SET_BACKEND, _BACKEND
    # Check whether the user has modified the backend.
    mod = inspect.getmodule(inspect.stack()[1][0])
    if 'agml.' not in mod.__name__:
        _USER_SET_BACKEND = True

    # If the backend is the same, don't do anything.
    if backend == _BACKEND:
        return

    _check_tf_torch()
    if backend not in ['tensorflow', 'tf', 'torch', 'pytorch']:
        raise ValueError(f"Invalid backend: {backend}.")
    if backend in ['tensorflow', 'tf'] and _BACKEND != 'tensorflow':
        if not _HAS_TENSORFLOW:
            raise ImportError(
                "TensorFlow not found on system, cannot be used as "
                "backend. Try running `pip install tensorflow`.")
        _BACKEND = 'tf'
        log("Switched backend to TensorFlow.", level = logging.INFO)
    elif backend in ['torch', 'pytorch'] and _BACKEND != 'torch':
        if not _HAS_TORCH:
            raise ImportError(
                "PyTorch not found on system, cannot be used as "
                "backend. Try running `pip install torch`.")
        _BACKEND = 'torch'
        log("Switched backend to PyTorch.", level = logging.INFO)


def user_changed_backend():
    """Returns whether the backend has been manually changed."""
    return _USER_SET_BACKEND


# Ported from https://github.com/tensorflow/tensorflow/blob/master/tensorflow/python/util/lazy_loader.py
class LazyLoader(types.ModuleType):
  """Lazily import a module, mainly to avoid pulling in large dependencies.  """
  def __init__(self, local_name, parent_module_globals, name):
    self._local_name = local_name
    self._parent_module_globals = parent_module_globals
    super(LazyLoader, self).__init__(name)

  def _load(self):
    """Load the module and insert it into the parent's globals."""
    # Import the target module and insert it into the parent's namespace.
    module = importlib.import_module(self.__name__)
    self._parent_module_globals[self._local_name] = module

    # Update this object's dict so that if someone keeps a reference to the
    #   LazyLoader, lookups are efficient (__getattr__ is only called on lookups
    #   that fail).
    self.__dict__.update(module.__dict__)
    return module

  def __getattr__(self, item):
    module = self._load()
    return getattr(module, item)

  def __dir__(self):
    module = self._load()
    return dir(module)


# Load TensorFlow and PyTorch lazily to prevent pulling them in when unnecessary.
torch = LazyLoader('torch', globals(), 'torch')
torch_data = LazyLoader('torch_data', globals(), 'torch.utils.data')
torchvision = LazyLoader('torchvision', globals(), 'torchvision')
tf = LazyLoader('tensorflow', globals(), 'tensorflow')


######### GENERAL METHODS #########

def _convert_image_to_torch(image):
    """Converts an image (np.ndarray) to a torch Tensor."""
    if isinstance(image, (list, tuple)):
        return torch.tensor(image)
    if isinstance(image, torch.Tensor) or image.ndim == 4:
        if image.shape[0] == 1 and image.shape[-1] <= 3 and image.ndim == 4:
            return torch.from_numpy(image).permute(0, 3, 1, 2).float()
        return image
    if image.shape[0] > image.shape[-1]:
        return torch.from_numpy(image).permute(2, 0, 1).float()
    return torch.from_numpy(image)


def _postprocess_torch_annotation(image):
    """Post-processes a spatially augmented torch annotation."""
    try:
        if image.dtype.is_floating_point:
            image = (image * 255).int()
    except AttributeError:
        pass
    return image


def as_scalar(inp):
    """Converts an input value to a scalar."""
    if isinstance(inp, (int, float)):
        return inp
    if np.isscalar(inp):
        return inp.item()
    if isinstance(inp, np.ndarray):
        return inp.item()
    if isinstance(inp, torch.Tensor):
        return inp.item()
    if isinstance(inp, tf.Tensor):
        return inp.numpy()
    raise TypeError(f"Unsupported variable type {type(inp)}.")


def scalar_unpack(inp):
    """Unpacks a 1-d array into a list of scalars."""
    return [as_scalar(item) for item in inp]


def is_array_like(inp):
    """Determines if an input is a np.ndarray, torch.Tensor, or tf.Tensor."""
    if isinstance(inp, (list, tuple)): # no need to import tensorflow for this
        return False
    if isinstance(inp, np.ndarray):
        return True
    if isinstance(inp, torch.Tensor):
        return True
    if isinstance(inp, tf.Tensor):
        return True
    return False


def convert_to_batch(images):
    """Converts a set of images to a batch."""
    # If `images` is already an array type, nothing to do.
    if is_array_like(images):
        return images

    # NumPy Arrays.
    if isinstance(images[0], np.ndarray):
        if not consistent_shapes(images):
            images = np.array(images, dtype = object)
            log("Created a batch of images with different "
                "shapes. If you want the shapes to be consistent, "
                "run `loader.resize_images('auto')`.")
        else:
            images = np.array(images)
        return images

    # Torch Tensors.
    if isinstance(images[0], torch.Tensor):
        if not consistent_shapes(images):
            images = [image.numpy() for image in images]
            images = np.array(images, dtype = object)
            log("Created a batch of images with different "
                "shapes. If you want the shapes to be consistent, "
                "run `loader.resize_images('auto')`.")
        else:
            images = torch.stack(images)
        return images

    # TensorFlow Tensors.
    if isinstance(images[0], tf.Tensor):
        if not consistent_shapes(images):
            images = tf.ragged.stack(images)
            log("Created a batch of images with different "
                "shapes. If you want the shapes to be consistent, "
                "run `loader.resize_images('auto')`.")
        else:
            images = tf.stack(images)
        return images


######### AGMLDATALOADER METHODS #########

class AgMLObject(object):
    """Base class for the `AgMLDataLoader` to enable inheritance.

    This class solves a bug which arises when trying to dynamically
    inherit from `tf.keras.utils.Sequence` and/or `torch.utils.data.Dataset`.
    The fact that the `AgMLDataLoader` has this `AgMLObject` as a subclass
    enables it to be able to handle dynamic inheritance. This is the sole
    purpose of this subclass, it does not have any features.
    """


def _add_dataset_to_mro(inst, mode):
    """Adds the relevant backend class to the `AgMLDataLoader` MRO.

    This allows for the loader to dynamically inherent from the
    `tf.keras.utils.Sequence` and `torch.utils.data.Dataset`.
    """
    if mode == 'tf':
        if not get_backend() == 'tf':
            if user_changed_backend():
                raise StrictBackendError(change = 'tf', obj = inst)
            set_backend('tf')
        if tf.keras.utils.Sequence not in inst.__class__.__bases__:
            inst.__class__.__bases__ += (tf.keras.utils.Sequence, )
    if mode == 'torch':
        if not get_backend() == 'torch':
            if user_changed_backend():
                raise StrictBackendError(change = 'torch', obj = inst)
        if torch_data.Dataset not in inst.__class__.__bases__:
            inst.__class__.__bases__ += (torch_data.Dataset,)




# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import sys
import math
import json
import shutil
import functools
import subprocess as sp
from datetime import datetime as dt

from agml.utils.io import recursive_dirname


# If this file is imported multiple times, only run the check once.
_HELIOS_CHECK_DONE_IN_SESSION = False

# Paths to the Helios module and the relevant C++ files.
HELIOS_PATH = os.path.join(recursive_dirname(__file__, 2), '_helios/Helios')
CANOPY_SOURCE = os.path.join(
    HELIOS_PATH, 'plugins/canopygenerator/src/CanopyGenerator.cpp')
CANOPY_HEADER = os.path.join(
    HELIOS_PATH, 'plugins/canopygenerator/include/CanopyGenerator.h')
LIDAR_SOURCE = os.path.join(HELIOS_PATH, 'plugins/lidar/src/LiDAR.cpp')

# Path to the stored Helios configuration JSON file in root dir.
HELIOS_CONFIG_FILE = os.path.expanduser('~/.agml/helios_config.json')


@functools.lru_cache(maxsize = None)
def load_default_helios_configuration():
    """Loads the default Helios parameter configuration."""
    with open(HELIOS_CONFIG_FILE, 'r') as f:
        return json.load(f)


def reinstall_helios():
    """Re-installs the Helios module from scratch.

    See `_check_helios_installation()` for a specific description of what
    this method does. Running this method manually also clears out the
    Helios directory, which is why extra confirmation is required.
    """
    check = input("Are you sure you want to re-install Helios? [y|n] ")
    if check != 'y':
        print("Aborting re-installation")
    else:
        print("Re-installing Helios.")
    if os.path.exists(HELIOS_PATH):
        shutil.rmtree(HELIOS_PATH)
    _check_helios_installation()


# Run the Helios configuration check.
def _check_helios_installation():
    """Checks for the latest Helios installation (and does as such).

    If no existing Helios installation is found, then this runs a fresh
    installation of Helios. If Helios is already installed, then this
    checks if a new possible version exists, and then installs if so.

    Note that this check if only run a maximum of once per day (specifically,
    the check is only run if it has not been run in the last 48 hours). This
    is to prevent constant resource-consuming checks for Git updates.
    """
    # Update the global Helios check.
    global _HELIOS_CHECK_DONE_IN_SESSION
    _HELIOS_CHECK_DONE_IN_SESSION = True

    # Get the path to the Helios installation file.
    helios_file = os.path.join(
        os.path.dirname(os.path.dirname(__file__)),
        '_helios/helios_install.sh')
    helios_dir = os.path.join(
        os.path.dirname(helios_file), 'Helios')

    # Run the installation/update. If the Helios directory is not found,
    # then run a clean installation. Otherwise, check if a new version of
    # Helios is available, then update the existing repository.
    if not os.path.exists(helios_dir):
        sys.stderr.write("Existing installation of Helios not found.\n")
        sys.stderr.write("Installing Helios to:\n")
        sys.stderr.write("\t" + helios_dir + "\n")

        # Update the last check of the Git update.
        with open(os.path.expanduser('~/.agml/config.json')) as f:
            contents = json.load(f)
        with open(os.path.expanduser('~/.agml/config.json'), 'w') as f:
            contents['last_helios_check'] = dt.now().strftime("%B %d %Y %H:%M:%S")
            json.dump(contents, f, indent = 4)

    else:
        # Check if the Git update check has been run in the last 48 hours.
        with open(os.path.expanduser('~/.agml/config.json')) as f:
            contents = json.load(f)
        last_check = contents.get(
            'last_helios_check', dt(1970, 1, 1).strftime("%B %d %Y %H:%M:%S"))
        last_check = dt.strptime(last_check, "%B %d %Y %H:%M:%S")

        # If the last check has been run less than 48 hours ago, then
        # return without having updated Helios.
        if (dt.now() - last_check).days < 2:
            return

        # Check if there is a new version available.
        sys.stderr.write(
            f"Last check for Helios update: over {(dt.now() - last_check).days} "
            f"day(s) ago. Checking for update.\n")

    # Execute the installation/update script.
    process = sp.Popen(['bash', helios_file],
                       stdout = sp.PIPE, universal_newlines = True)
    for line in iter(process.stdout.readline, ""):
        sys.stderr.write(line)
    process.stdout.close()

    # Update the Helios parameters.
    sys.stderr.write("Updating Helios parameter configuration.")
    _update_helios_parameters()

    # Copy the project files to the Helios directory.
    project_dir = os.path.join(os.path.dirname(__file__), 'synthetic_data_generation')
    output_dir = os.path.join(helios_dir, 'projects', 'SyntheticImageAnnotation')
    shutil.copytree(project_dir, output_dir)
    os.makedirs(os.path.join(output_dir, 'xml'), exist_ok = True)

    # Recompile Helios.
    from agml.synthetic.compilation import _compile_helios_default
    _compile_helios_default()

    # Update the latest check.
    with open(os.path.expanduser('~/.agml/config.json')) as f:
        contents = json.load(f)
    with open(os.path.expanduser('~/.agml/config.json'), 'w') as f:
        contents['last_helios_check'] = dt.now().strftime("%B %d %Y %H:%M:%S")
        json.dump(contents, f, indent = 4)

    # Clear the outputs.
    sys.stdout.write('\n')
    sys.stderr.write('\n')


def _update_helios_parameters():
    """Updates all default Helios parameters (canopy, camera, LiDAR).
    
    Every time a new version of Helios is installed or updated, the default
    parameters need to be read from the Helios source files and updated in
    the local configuration file: `~/.agml/helios_config.json`.
    """
    params = {
        'canopy': _get_canopy_params(),
        'camera': _get_camera_params(),
        'lidar': _get_lidar_params()
    }

    # Save the parameters to the configuration file.
    with open(HELIOS_CONFIG_FILE, 'w') as f:
        json.dump(params, f, indent = 4)


def _get_canopy_params():
    """Updates the default canopy generation parameters for Helios."""
    # Generate the canopy types.
    with open(CANOPY_HEADER, 'r') as f:
        canopy_header_lines = f.readlines()
    canopy_types = [
        re.match('struct\\s(.*?){\\n', string).group(1).split('Parameters')[0]
        for string in canopy_header_lines if 'struct ' in string]

    # Generate the canopy parameters.
    with open(CANOPY_SOURCE) as f:
        canopy_source_lines = f.read().replace('\n', '')

    # Add the default parameters for ground annotations.
    canopy_parameters = {'Ground': {
        'origin': [0.0, 0.0, 0.0],
        'extent': [10.0, 10.0],
        'texture_subtiles': [10.0, 10.0],
        'texture_subpatches': [1.0, 1.0],
        'ground_texture_file': 'plugins/canopygenerator/textures/dirt.jpg',
        'rotation': 0.0}}

    # This regex is adapted from the following source:
    # https://stackoverflow.com/questions/943391/how-to-get-the-function-declaration-or-definitions-using-regex
    # It searches for all method definitions and then gets all of the content in the method.
    for content in re.finditer(
            r"((?<=[\s:~])(\w+)\s*\(([\w\s,<>\[\].=&':/*]*?)\)"
            r"\s*(const)?\s*(?={)){(.*?)}", canopy_source_lines):
        # Ensure that we're only checking valid parameter methods.
        name = content.group(2).split('Parameters')[0]
        source = content.group(5)

        # Get all of the definitions by removing all spaces of at least length 2.
        if any(i == name for i in canopy_types) and 'std::cout' not in source:  # noqa
            canopy_parameters[name] = {}
            definitions = [
                s.replace(';', '') for s in re.split('[\\s]{2,}', source)
                if s != '' and not s.startswith('//')]
            for d in definitions:
                # Ignore comments.
                if d.startswith('//'):
                    continue

                # Split the name and the value.
                n, value = d.split(' = ')

                # If the value is defined in a method, extract the method arguments.
                if '(' in value:
                    value = re.search('\\((.*?)\\)', value).group(1)

                # Remove unnecessary terms. The `.f` removal needs to be done using
                # regex since we can't just remove regular `f` in case of paths,
                # but we also can't just replace `.f` with `0` or `.0`.
                value = value.replace(',', ' ').replace('\"', '')
                value = re.sub('(\\.|\\d)f', r'\g<1>0', value)
                value = re.sub('//.*?$', '', value)

                # A specific case to replace values using pi.
                value = value.strip().replace('  ', ' ')
                if 'M_PI' in value:
                    value = round(float(value.split('*')[0].strip()) * math.pi, 6)

                # Convert numerical values to numbers/lists, as necessary.
                elif value.isnumeric():
                    value = int(value)
                elif value.replace('.', '').isdigit():
                    value = float(value)
                elif ' ' in value:
                    value_items = value.split(' ')
                    if all(i.isdigit() for i in value_items):
                        value = [int(v) for v in value_items]
                    else:
                        value = [float(v) for v in value_items]

                # If the value is a path, then make it absolute.
                else:
                    if value.endswith('.jpg') or value.endswith('.png'):
                        value = os.path.join(HELIOS_PATH, value)

                # Update the parameter dictionary.
                canopy_parameters[name][n] = value

    # Return the two dictionaries.
    return {'types': canopy_types, 'parameters': canopy_parameters}


def _get_camera_params():
    """Updates the default camera parameters for Helios."""
    # No files are read here, the parameters are hard-coded.
    camera_params = {
        'image_resolution': [600, 400],
        'camera_lookat': [0.0, 0.0, 1.0],
        'camera_position': [0.0, -2.0, 1.0]}
    return {'parameters': camera_params}


def _get_lidar_params():
    """Updates the default LiDAR parameters for Helios."""
    # No files are read here, the parameters are hard-coded.
    lidar_params = {
        "origin": [0.0, 0.0, 0.0],
        "size": [250.0, 450.0],
        "thetaMin": 0,
        "thetaMax": 180,
        "phiMin": 0,
        "phiMax": 360,
        "exitDiameter": 0.0,
        "beamDivergence": 0.0,
        "ASCII_format": "x y z"}
    return {'parameters': lidar_params}


################## API FUNCTIONS FOR CONVENIENCE ##################


def available_canopies():
    """Returns a list of available canopies in Helios."""
    return load_default_helios_configuration()['canopy']['types']


def default_canopy_parameters(canopy):
    """Returns the default canopy parameters for a given `canopy`."""
    return load_default_helios_configuration()['canopy']['parameters'][canopy]




# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__version__ = '0.4.0'
__all__ = ['data', 'backend', 'viz']


# If AgML is being imported for the first time, then we need to setup
# the module, namely prepping the config file.
def _setup():
    import os as _os
    import json as _json
    if not _os.path.exists(_os.path.expanduser('~/.agml')):
        _os.makedirs(_os.path.expanduser('~/.agml'))
        with open(_os.path.join(
                _os.path.expanduser('~/.agml/config.json')), 'w') as f:
            _json.dump({'data_path': _os.path.expanduser('~/.agml/datasets')}, f)
_setup(); del _setup # noqa


# There are no top-level imported functions or classes, only the modules.
from . import data, backend, synthetic, viz







# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import time
import inspect
import logging


# This is a simple hack to auto-adjust `tqdm` based on whether
# we are in a Jupyter notebook or in a regular shell environment.
if 'get_ipython' in globals():
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm


# Track already given warnings.
_GIVEN_WARNINGS = {}


# Configure the logger.
logging.basicConfig(
    format = '%(asctime)s %(levelname)s - %(name)s: %(message)s',
    datefmt = '%m-%d-%Y %H:%M:%S')


def log(msg, level = logging.WARNING):
    """Logs a message to the console.

    This is an internal wrapper method for the logging module,
    that takes care of the different logging warning levels
    and ensures that no log is repeated twice at the same time.
    """
    method_map = {
        logging.DEBUG: logging.debug,
        logging.INFO: logging.info,
        logging.WARNING: logging.warning,
        logging.CRITICAL: logging.critical,
        'debug': logging.debug,
        'info': logging.info,
        'warning': logging.warning,
        'critical': logging.critical
    }

    # Check if the warning has been delivered recently.
    log_method = method_map[level]
    save_hash = f"{round(time.time(), -1)}:{inspect.stack()[1].function}:{msg}"
    msg = "[AgML] " + msg
    if save_hash in _GIVEN_WARNINGS.keys():
        if _GIVEN_WARNINGS[save_hash] == msg:
            save_time = float(save_hash.split(':')[0])
            if time.time() - save_time > 240:
                log_method(msg)
    else:
        log_method(msg)
    _GIVEN_WARNINGS[save_hash] = msg


class no_print(object):
    """Doesn't print anything when this is used.

    This should be used in a `with` statement for context, namely

    > with no_print():
    >   print("This won't be printed.")
    > print("This will be printed.")
    This will be printed.
    """
    def __enter__(self):
        self._stdout_reset = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._stdout_reset




# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy


class AgMLSerializable(object):
    """Base class for all AgML serializable objects.

    Most objects in AgML inherit from this class, and are thus
    serializable into a JSON or binary (such as pickle) framework,
    allowing objects to be easily accessed, adjusted, and copied.

    More specifically, subclasses of this will have an automatically
    defined `__getstate__` and `__setstate__` method which consists
    of a dictionary with the necessary class attributes. In addition,
    it will have a `__copy__` and `__deepcopy__` method which have
    the same behavior (the class is always deep copied).

    Subclasses only need to define a `serializable` property with a
    frozen set of strings containing the attributes that are to be
    serialized. The expectation is that the strings in the set will
    all be the name of attributes minus a leading underscore. If
    there are parameters which are just the name of the attribute,
    no leading underscore, also add them to another property of the
    class, `state_override`, to ensure compatibility.

    In turn, objects can be used with a JSON serialization format,
    the pickle serialization format, or copied as desired.
    """
    serializable: "frozenset"
    state_override: "frozenset"

    def __init_subclass__(cls, **kwargs):
        if not hasattr(cls, 'state_override'):
            cls.state_override = frozenset(())

    def __getstate__(self):
        state = {}
        for param in self.serializable:
            try:
                state[param] = getattr(self, f'_{param}')
            except AttributeError:
                if param in self.state_override:
                    state[param] = getattr(self, param)
                else:
                    raise AttributeError(
                        f"Encountered error while attempting to serialize "
                        f"a {self.__class__}: the attribute '_{param}' "
                        f"(or '{param}') does not exist.")
        return state

    def __setstate__(self, state):
        for field in state.keys():
            if field in self.state_override:
                setattr(self, field, state[field])
            else:
                setattr(self, f'_{field}', state[field])

    def __deepcopy__(self, memo = None):
        params = self.__getstate__()
        cls = super(AgMLSerializable, self).__new__(self.__class__)
        cls.__setstate__(copy.deepcopy(params))
        return cls

    def __copy__(self):
        return self.__deepcopy__()







# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .tools import (
    format_image,
    set_colormap,
    get_colormap
)
from .masks import (
    output_to_mask,
    visualize_image_and_mask,
    overlay_segmentation_masks,
    visualize_overlaid_masks,
    visualize_image_mask_and_predicted
)
from .boxes import (
    annotate_bboxes_on_image,
    visualize_image_and_boxes,
    visualize_real_and_predicted_bboxes,
    visualize_image_and_many_boxes
)
from .labels import (
    visualize_images,
    visualize_images_with_labels
)
from .inspection import (
    plot_synthetic_camera_positions,
    visualize_all_views
)



# Copyright 2021 UC Davis Plant AI and Biophysics Lab
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import cv2


def consistent_shapes(objects):
    """Determines whether the shapes of the objects are consistent."""
    try:
        shapes = [i.shape for i in objects]
        return shapes.count(shapes[0]) == len(shapes)
    except:
        try:
            lens = [len(i) for i in objects]
            return lens.count(lens[0]) == len(lens)
        except:
            if isinstance(objects[0], (int, float)):
                return True
            return False


def needs_batch_dim(image):
    """Determines whether an image has or is missing a batch dimension."""
    if not hasattr(image, 'shape'):
        raise TypeError(
            "Can only determine batch dimensions for numpy arrays or tensors.")
    if len(image.shape) == 2:
        return True
    elif len(image.shape) == 3:
        if image.shape[0] != 1:
            return True
    return False


def resolve_image_size(size):
    if isinstance(size, int):
        return size, size
    elif isinstance(size, (list, tuple, set, np.ndarray)):
        if not len(size) == 2:
            raise ValueError(
                "Only two values must be provided for an input image size.")
        return size
    else:
        raise ValueError(
            f"Expected either an integer or list of two values for "
            f"image size, got ({size}) of type ({type(size)}).")


class imread_context(object):
    """Wraps the `cv2.imread` function into a context.

    This context allows for a better error message regarding the
    loading of images using OpenCV. Traditionally, if the path
    doesn't exist or the image is empty, the `imread` function
    will just return `None`, which will usually lead to a cryptic
    error message in future operations. This context catches such
    issues and raises a more detailed error message for the user.
    """
    def __init__(self, path, flags = None):
        self._path = path
        if not os.path.exists(path):
            raise ValueError(
                f"The image path '{path}' does not exist. "
                f"Please check the dataset you are using, "
                f"and if files are missing, re-download it.")
        self.flags = flags
        if self.flags is None:
            self.flags = cv2.IMREAD_UNCHANGED

    def __enter__(self):
        try:

            img = cv2.imread(self._path, self.flags)
        except cv2.error:
            raise ValueError(
                f"An error was encountered when loading the image "
                f"'{self._path}'. Please check the dataset you are "
                f"using and re-download it if files are missing.")
        else:
            if img is None:
                raise ValueError(
                    f"The image at '{self._path}' is empty,"
                    f" corrupted, or could not be read. Please "
                    f"re-download the dataset you are using.")
        return img

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


